/*
 * Cell.h
 *
 * Copyright (C) 1999, Thierry Matthey <matthey@ii.uib.no>
 *
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#ifndef _CELL_H
#define _CELL_H

#include "List.h"
#include "FastList.h"
#include "Particle.h"
#include "CellIndex.h"

#ifdef HAVE_MPI
#ifdef HAVE_ONE
#include <mpi.h>
#endif
#endif

//_________________________________________________________________ Cell

class Cell {

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Constructors, destructors, assignment
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 public:
  Cell();
  Cell(CellIndex index,List<CellIndex>* path);
  ~Cell();

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// New methods of class Cell
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 public:
  void add(const Particle& item);
  FastList<Particle>* getParticles() const;
  void setPath(List<CellIndex>* path);
  List<CellIndex>* getPath() const;
  void setIndex(const CellIndex& index);
  const CellIndex& getIndex() const;
  void setOwnerId(int id);
  int  getOwnerId() const;
  void setActualId(int id);
  int  getActualId() const;
  void setWeight(int w);
  int  getWeight() const;
  long count() const;
  long size() const;
  bool own() const;
  bool have() const;
  bool home() const;
  Particle* getItems() const;
  Particle* setItems(long size);
  Particle& item(long index) const;
  void atLeastSize(long leastsize);
  void remove(long index);
  void usedByAdd(int id);
  void usedByReset();
  int  usedBy(long index) const;
  bool usedById(int id) const;
  bool usedByAny() const;
  void reduce();

#ifdef HAVE_MPI
#ifdef HAVE_ONE 
  void  setShmPtr(void* ptr) {myShmPtr = ptr;};
  void* getShmPtr() const {return myShmPtr;};
  void* getPtr() const {return myParticles->getPtr();};
#endif
#endif

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// My data members
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 private:
  int                  myUsedBy[max_neighbours];
  int                  myOwnerId;
  int                  myActualId;
  int                  myWeight;
  CellIndex            myIndex;
  FastList<Particle>*  myParticles;
  List<CellIndex>*     myPath;

#ifdef HAVE_MPI
#ifdef HAVE_ONE 
  void*                myShmPtr;
#endif
#endif
};

//______________________________________________________________________ INLINES

inline void Cell::reduce(){
  myParticles->reduce();
}
inline Particle* Cell::setItems(long size){
  return myParticles->setItems(size);
}

inline void Cell::setOwnerId(int id){
  myOwnerId = id;
}

inline int Cell::getOwnerId() const{
  return myOwnerId;
}

inline void Cell::setActualId(int id){
  myActualId = id;
}

inline int Cell::getActualId() const{
  return myActualId;
}
inline void Cell::setWeight(int w){
  myWeight = w;
}

inline int Cell::getWeight() const{
  return myWeight;
}

inline List<CellIndex>* Cell::getPath() const{
  return myPath;
}

inline FastList<Particle>* Cell::getParticles() const{
  return myParticles;
}

inline long Cell::count() const{
  return myParticles->count();
} 

inline long Cell::size() const{
  return myParticles->size();
} 

inline Particle* Cell::getItems() const{
  return myParticles->getItems();
}

inline Particle& Cell::item(long index) const{
  return myParticles->item(index);
}

inline void Cell::remove(long index){
  myParticles->remove(index);
} 

inline void Cell::setIndex(const CellIndex& index){
  myIndex = index;
}

inline const CellIndex& Cell::getIndex() const{
  return myIndex;
}

inline void Cell::add(const Particle& item){
  myParticles->add(item);
}

inline bool Cell::own() const{
  return (myOwnerId == myId);

}
inline bool Cell::have() const{
  return (myActualId == myId);
}

inline bool Cell::home() const{
  return (myActualId == myId && myOwnerId == myId);
}

inline int Cell::usedBy(long index) const{
  return myUsedBy[index];
}

inline bool Cell::usedByAny() const{
  return (myUsedBy[0]>=0);
}

inline void Cell::atLeastSize(long leastsize){
  myParticles->atLeastSize(leastsize);
}

#endif
















































