/*
 * Force.C
 *
 * Copyright (C) 1999-2000, Thierry Matthey <matthey@ii.uib.no>
 *
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#include "Force.h"

#include "Particle.h"
#include "mathutilities.h"

//_________________________________________________________________ Force

Force::Force(){
  fx       = 0.0;
  fy       = 0.0;
  myMin    =-1.0;
  myRCut   = 0.0;
  myRCutSq = 0.0;
}

Force* Force::copy() const
{
  return new Force(*this);
}

#ifdef HAVE_EPOT
bool Force::calc(Particle& a, Particle& b, Real& epot, Real bx, Real by){
#else
bool Force::calc(Particle& a, Particle& b, Real&, Real bx, Real by){
#endif

  Real rxij = a.x-b.x - bx;
  Real ryij = a.y-b.y - by;
  Real rij2 = rxij*rxij + ryij*ryij;
  
  if (rij2 <= myRCutSq){
#ifdef HAVE_FORCE126
    // We know that the exponents are 12 and 6
    long i = a.tag;
    long j = b.tag;
    Real rrij2 = 1.0/rij2;
    Real rrij6 = rrij2*rrij2*rrij2;
    Real fij = con4[i][j] * rrij6*(rrij6*(Real)nalf[i][j]*alfp[i][j]-(Real)nbet[i][j]*betp[i][j]) * rrij2;
#ifdef HAVE_EPOT
    Real ep = con4[i][j] * rrij6*(rrij6*alfp[i][j]-betp[i][j])-epsilon[i][j];
#endif
#else
    Real rij = sqrt(rij2);
    long i = a.tag;
    long j = b.tag;
    long mal = nalf[i][j];
    long nal = nbet[i][j];
    Real u = alf[i][j]/rij;
    Real v = bet[i][j]/rij;
    Real fij = con4[i][j] * ((Real)mal*power(u,mal)-(Real)nal*power(v,nal))/rij2;
#ifdef HAVE_EPOT
    Real ep = con4[i][j] * (power(u,mal)-power(v,nal));
#endif
#endif
    Real fxij = fij * rxij;
    Real fyij = fij * ryij;
    //cerr <<myId<<": ("<<a.i<<","<<b.i<<"): "<<fxij<<","<<fyij<<"\n";
    a.fxa += fxij;
    a.fya += fyij;
    b.fxa -= fxij;
    b.fya -= fyij;
#ifdef HAVE_EPOT
    epot   += ep;
    a.epot += ep;
    b.epot += ep;
#endif
    return true;
  }
  else
    return false;
}

void Force::readParameters(ifstream& parameters){
  // Read the force parameters for the particles
  long n,i,k,l;

  for(l=0;l<max_NTypes;l++)
    for(k=0;k<max_NTypes;k++)
      used[l][k] = false;

  parameters >> n;
  nextLine(parameters);

  for(i=0;i<n;i++){
    parameters >> l >> k;
    nextLine(parameters);
    l--; // FORTRAN
    k--; // FORTRAN

    parameters >> con[l][k] >> alf[l][k]  >> nalf[l][k] >> bet[l][k] >> nbet[l][k];
    nextLine(parameters);
    used[l][k] = true;
  }

  // Fill up the rest of the table if needed
  for(l=0;l<max_NTypes;l++){
    for(k=0;k<max_NTypes;k++){
      if (used[l][k] && !used[k][l]){
        used[k][l] = used[l][k];
        con[k][l]  = con[l][k];
        alf[k][l]  = alf[l][k];
        nalf[k][l] = nalf[l][k];
        bet[k][l]  = bet[l][k];
        nbet[k][l] = nbet[l][k];
      }
    }
  }
  // Do some precomputing ...
  for(l=0;l<max_NTypes;l++){
    for(k=0;k<max_NTypes;k++){
      con4[l][k] = 4*con[l][k];
#ifdef HAVE_FORCE126
      nalf[l][k] = 12;
      nbet[l][k] = 6;
#endif
      alfp[l][k] = power(alf[l][k],nalf[l][k]);
      betp[l][k] = power(bet[l][k],nbet[l][k]);
      if(used[l][k]){
	zero[l][k] = pow((Real)nalf[l][k]/(Real)nbet[l][k]*alfp[l][k]/betp[l][k],1.0/(Real)(nalf[l][k]-nbet[l][k]));      
	if(myMin < 0.0 || myMin > zero[l][k])
	  myMin = zero[l][k];
      }

      if(used[l][k] && myId == master)
	cerr << "R0("<<l<<","<<k<<") = " << zero[l][k] <<"\n";
    }
  }
}

void Force::nextLine(ifstream& in){
  char a;
  while(in.get(a))
    if(a == '\n')
      break;
}

void Force::setRCut(Real rcut){
  myRCut = rcut;
  myRCutSq = rcut*rcut;
  for(long l=0;l<max_NTypes;l++){
    for(long k=0;k<max_NTypes;k++){
      if(used[l][k]){
	epsilon[l][k] = con4[l][k] * (power(alf[l][k]/rcut,nalf[l][k])-power(bet[l][k]/rcut,nbet[l][k]));
      if(myId == master)
	cerr << "epsilon("<<l<<","<<k<<") = " << epsilon[l][k] <<"\n";
      }
    }
  }
}
