/*
 * StartWorld.C
 *
 * Copyright (C) 1999-2000, Thierry Matthey <matthey@ii.uib.no>
 *
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#include "StartWorld.h"

#ifdef HAVE_MPI
#include <mpi.h>
#endif

#include "global.h"
#include "Universe.h"
#include "Force.h"
#include "mathutilities.h"

//_________________________________________________________________ StartWorld

#ifdef HAVE_MPI
static const long package_size = 1000;
//static const long reading_size = 100000;
#endif

StartWorld::StartWorld(char *filename, Force* force){

  long i;
  Real x,y;

  myForce = force;
  myOutFilename[0] = '\0';
  myInFilename[0] = '\0';
  //cerr << "Reading from parameterfile '"<<filename<<"'.\n";

  ifstream parameters(filename, ios::in);
  if(!parameters){
    if(myId == master)
      cerr << "Can not open parameterfile '"<<filename <<"'.\n";
    myNStop = -1;
    return;
  }

  // The outfile
  parameters >> myOutFilename;
  nextLine(parameters);

  // The inoutfile of the particles
  parameters >> myInFilename;
  nextLine(parameters);

  ifstream input(myInFilename, ios::in);
  if(!input){
    if(myId == master)
      cerr << "Can not open inputfile '"<< myInFilename <<"' for particles.\n";
    myNStop = -1;
    return;
  }

  // tstop, dt, nstop
  parameters >> myTStop >> myDt >> myNStop;
  nextLine(parameters);

  // Time interval for output of energy
  // Number of output levels
  parameters >> myDj;
  nextLine(parameters);

  // RCut - cell size
  parameters >> myRCut;
  nextLine(parameters);

  // Temperature
  parameters >> myTemp;
  nextLine(parameters);

  // Number of atomic types
  parameters >> myNTypes;
  nextLine(parameters);
  //cerr << "NTypes : "<< myNTypes<<"\n";

  // Masse and percentage of the types
  myMaxRadius = -1.0;
  for(i=0;i<myNTypes;i++){
    parameters >> myMass[i] >> myProcent[i] >> myRadius[i];
    if(myRadius[i] <= 0.0){
      myRadius[i] = 1.0;
      if(myId == master)
	cerr <<"Radius negativ or missing.\n";
    }
    if(myRadius [i] > myMaxRadius)
      myMaxRadius = myRadius[i];

    nextLine(parameters);
    //cerr << myMass[i] <<","<< myProcent[i]<<","<< myRadius[i]<<"\n";
  }

  // fx, fy (constant force)
  parameters >> force->fx >> force->fy;
  nextLine(parameters);

  // The Force object knows best how to read ...
  force->readParameters(parameters);

  // Compute the size of the slap based on the 
  // position of the particles
  myRight = myLeft;
  myTop = myBottom;

  myNumber = 0;

  if(myId == master){
    while(input >> x >> y){
      myNumber++;
      if (x < myLeft) myLeft = x;
      if (x > myRight) myRight = x;
      if (y < myBottom) myBottom = y;
      if (y > myTop) myTop = y;
    }
  }
  input.close();
#ifdef HAVE_MPI
  MPI_Bcast(&myNumber,1, MPI_LONG, master,MPI_COMM_WORLD);
  MPI_Bcast(&myLeft,1, TYPE_REAL, master,MPI_COMM_WORLD);
  MPI_Bcast(&myRight,1, TYPE_REAL, master,MPI_COMM_WORLD);
  MPI_Bcast(&myBottom,1, TYPE_REAL, master,MPI_COMM_WORLD);
  MPI_Bcast(&myTop,1, TYPE_REAL, master,MPI_COMM_WORLD);
#endif  
  myReadLeft   = myLeft;
  myReadRight  = myRight;
  myReadTop    = myTop;
  myReadBottom = myBottom;

  myLeft   -= myMaxRadius;
  myRight  += myMaxRadius;
  myBottom -= myMaxRadius;
  myTop    += myMaxRadius;

  // Pull
  if(myNTypes>3){
    parameters >> myPull >> myPullStart >> myPullStop;
    nextLine(parameters);

    myBottom -= 2*myRadius[myNTypes-1]+fabs(myPull)*(myPullStop-myPullStart);
    myTop    += 2*myRadius[myNTypes-1]+fabs(myPull)*(myPullStop-myPullStart);
    if(myId == master)
      cerr <<"Using pull particles with v = "<<myPull<<", t = ("<<myPullStart<<","<<myPullStop<<").\n";

  }

  parameters.close();

  // Print statistic
  if(myId == master){
    cerr << "Read "<<myNumber<<" particles.\n";  
    cerr << "Input: "<<myInFilename <<"\n";
    cerr << "Output: "<<myOutFilename <<"\n";
    cerr << "Area: (" << myLeft <<","<< myBottom <<")-("<< myRight <<","<< myTop <<"), (" << myRight-myLeft <<","<< myTop-myBottom <<")\n";
    cerr << "Rcut: "<<myRCut<<"\n";
    cerr << "Max particle radius: "<<myMaxRadius<<"\n";
  }
}

StartWorld::~StartWorld(){
  delete myForce;
}

void StartWorld::addParticles(Universe* universe){


  long l = 0;
  long k = 0;
  long j = 0;
  Real x,y,u,v;
  long i;
  long* tmpInt = new long[numProc]; 
    for(i=0;i<numProc;i++)
      tmpInt[i] =0;

  if(myId == master) {
    ifstream input(myInFilename, ios::in);
    FastList<Particle>* tmpParticles = new FastList<Particle>[numProc];
    while(input >> x >> y){

      nextLine(input);
      Real t = myrandom()*100;
      Real s = 0.0;
      int id = universe->getOwnerId(x,y);
      tmpInt[id]++;
      for(i=0;i<myNTypes;i++){
	if(s <= t && t < s+ myProcent[i]) {
	  u = gauss()*sqrt(K_BOLTZMANN*getTemp()/myMass[i]);
	  v = gauss()*sqrt(K_BOLTZMANN*getTemp()/myMass[i]);
	  Particle par(Particle(x,y,u,v,myRadius[i],myMass[i],(short)i,j));
	  if(myId == id || numProc < 2){
	    universe->addParticle(par);
	    l++;
	  }
#ifdef HAVE_MPI
	  else{
	    tmpParticles[id].add(par);
	    if(tmpParticles[id].count() > package_size){
	      MPI_Send(tmpParticles[id].getItems(),(int)tmpParticles[id].count(),TYPE_PARTICLE,id,UPDATE_CELLS,MPI_COMM_WORLD);
	      tmpParticles[id].removeAll();
	    }
	  }
#endif
	  break;
	}
	s += myProcent[i];
      }      
      j++;
    }

    input.close();
#ifdef HAVE_MPI
    for(i=0;i<numProc;i++){
      if(i != myId){
	if(tmpParticles[i].count() > 0){
	  MPI_Send(tmpParticles[i].getItems(),(int)tmpParticles[i].count(),TYPE_PARTICLE,(int)i,UPDATE_CELLS,MPI_COMM_WORLD);
	  tmpParticles[i].removeAll();
	}
	MPI_Send(tmpParticles[i].getItems(),(int)tmpParticles[i].count(),TYPE_PARTICLE,(int)i,UPDATE_CELLS,MPI_COMM_WORLD);
      }
    }
#endif    
    delete [] tmpParticles;
  }
#ifdef HAVE_MPI
  else {
    FastList<Particle>* tmpParticles = new FastList<Particle>(package_size);


    int count = 0;
    l = 0;

    do {
      MPI_Status status;
      MPI_Probe(MPI_ANY_SOURCE,UPDATE_CELLS,MPI_COMM_WORLD, &status);
      MPI_Get_count(&status,TYPE_PARTICLE,&count);
      MPI_Recv(tmpParticles->setItems(count),count,TYPE_PARTICLE,master,UPDATE_CELLS,MPI_COMM_WORLD, &status);
      for(long m=0;m<tmpParticles->count();m++)
	universe->addParticle(Particle(tmpParticles->item(m)));
      l += count;
    }
    while (count > 0);

    delete tmpParticles;
     
  }
  MPI_Bcast(tmpInt,numProc, MPI_LONG, master,MPI_COMM_WORLD);
#endif
     
     

#ifdef HVAE_PULL
  if(myNTypes>3){
    long tag = myNTypes-1;
    long n = (int)(universe->getLx()/(2*myRadius[tag]));
    if(myId == master)
      cerr << "Adding "<<2*n<<" pull particles.\n"; 
    Real y1 = myReadBottom - 2.75003;//3*myRadius[tag];
    Real y2 = myReadTop + 2.75003;//3*myRadius[tag];
    universe->setL0(y2-y1);
    u = 0;
    v = myPull;
    for(long i=0;i<n;i++){
      x = universe->getLeft()+((Real)i+0.5)*universe->getLx()/(Real)n;
      if(myId == universe->getOwnerId(x,y1)){
	universe->addParticle(Particle(x,y1,u,-v,myRadius[tag],myMass[tag],(short)tag,j));
	l++;
	k++;
      }
      if(myId == universe->getOwnerId(x,y2)){
	universe->addParticle(Particle(x,y2,u,v,myRadius[tag],myMass[tag],(short)tag,j));
	l++;
	k++;
      }
      j++;
    }
  }
#endif
  for(j=0;j<numProc;j++){
#ifdef HAVE_MPI
      MPI_Barrier(MPI_COMM_WORLD);
#endif
      if(myId==j){
	cerr <<"Proc "<<myId<<" has "<<l<<" ("<<tmpInt[myId]<<","<<k<<") particles, read "<<myNumber<<".\n";
    }    
#ifdef HAVE_MPI
      MPI_Barrier(MPI_COMM_WORLD);
#endif
  }

  myNumber = l;
  delete [] tmpInt;
}

void StartWorld::nextLine(ifstream& in){
  char a;
  while(in.get(a))
    if(a == '\n')
      break;
}

