/*
 * Universe.C
 *
 * Copyright (C) 1999-2000, Thierry Matthey <matthey@ii.uib.no>
 *
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#include <stdlib.h>
#include <stdio.h>
 
#ifdef HAVE_MPI
#include <mpi.h>
#endif
#ifdef HAVE_METIS
#include "metis.h"
#endif

#include "Universe.h"
#include "StartWorld.h"
#include "Force.h"
#include "Particle.h"
#include "mathutilities.h"

//_________________________________________________________________ Universe

ofstream output;
static const long stat_number = 100;

#ifdef HAVE_MPI
// Deadlock
static const double timeout_deadlock = 0.5;
static const long n1 = 5000; // number of cell indicies
static const long n2 = 100;  // number of cells
#endif

static char id_char[]={
  '0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z','!','"','#','$','%','&','\'','(',')','*','+',',','-','.','/','[','\\',']','^','_','`','{','|','}','~',':',';','<','=','>','?','@','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','',''
};

Universe::Universe(StartWorld* world, CommunicationType com, PartitionType part, bool dumpflag){

  // Test if StartWorld is ok ... otherwise abort
  if(world->getNStop()<1){
    myNStop       = -1;
    delete world;
    return;
  }

  long i,j,k,l;

  myCommunication   = com;
  myPartition       = part;
  myDump            = dumpflag;
  myForce           = NULL;
  myCells           = NULL;
  myOwnCells        = NULL;
  myOtherCells      = NULL;
  myOtherCellsUsed  = NULL;
  myMissingCells    = NULL;
  myTmpList         = NULL;
  myOverlap         = NULL;
  myNeighbours      = NULL;
  myRequestCell     = NULL;
  myRequestId       = NULL;
  myUpdateParticles = NULL;
  mySendList        = NULL;
  myRecvList        = NULL;
  myIntArray        = NULL;
  myIntArray2       = NULL;
  myDumpTab         = NULL;
  myBoundsX         = NULL;
  myBoundsY         = NULL;
#ifdef HAVE_MPI
  myRequestArray    = NULL;
  myStatusArray     = NULL;
  myRequestList     = NULL;
  myStatusList      = NULL;
  myBuffer          = NULL;
  myNBuffer         = 0;
#endif
  myCount           = 0;
  myTime            = 0.0;
  myDtCount         = 0;
  myL0              = 1.0;
  myL1              = 1.0;
  myDumpCount       = 0;
  // Statistic
  myTimeTotal       = 0.0;
  myTimeSim         = 0.0;
  myTimeCalc        = 0.0;
  myTimeCom         = 0.0;
  myTimeComPush     = 0.0;
  myTimeComPop      = 0.0;
  myTimeComGet      = 0.0;
  myTimeIdle        = 0.0;
  myTimeSim         = 0.0;
  myTimeUpdate      = 0.0;
  myTimeDist        = 0.0;
  myTimeRest        = 0.0;
  myCom             = 0;
  myComInter        = 0;
  myComBrute        = 0;
  myDistribute      = 0;
  myParInter        = 0;

  computeTimerOverhead();

  // Init and get datas from StartWorld
  myNStop       = world->getNStop();
  myForce       = world->getForce()->copy();
  myNStop       = world->getNStop();
  myTStop       = world->getTStop();
  myPullStart   = world->getPullStart();
  myPullStop    = world->getPullStop();
  myLeft        = world->getLeft()-world->getMaxRadius();
  myRight       = world->getRight()+world->getMaxRadius();
  myLx          = myRight - myLeft; 
  myBottom      = world->getBottom()-world->getMaxRadius();
  myTop         = world->getTop()+world->getMaxRadius();
  myLy          = myTop - myBottom;
  myDt          = world->getDt();
  myDj          = world->getDj();
  myRCut        = world->getRCut();
  myTemp        = world->getTemp();
  myN           = long((myLx)/myRCut);
  myM           = long((myLy)/myRCut);
  myNumberTotal = world->getNumber();
  // Add som cells around if non-periodic boundaries
  // to get better performance
  if (!periodic_boundary_x){
    myN     += 3;
    if(myCommunication == CommunicationOneSided || myCommunication == CommunicationOneSidedFast)
      myN     += 2;
    myLeft  -= ((Real)myN*myRCut-myLx)/2.0;
    myRight += ((Real)myN*myRCut-myLx)/2.0;
    myLx     = (Real)myN*myRCut;
  }
  if (!periodic_boundary_y){
    myM      += 3;
    if(myCommunication == CommunicationOneSided || myCommunication == CommunicationOneSidedFast)
      myM     += 2;
    myBottom -= ((Real)myM*myRCut-myLy)/2.0;
    myTop    += ((Real)myM*myRCut-myLy)/2.0;
    myLy      = (Real)myM*myRCut;
  }
  myA           = myLx/(Real)myN;
  myB           = myLy/(Real)myM;
  myDensity     = myA*myB/(sqrt(3.0)/2.0*myForce->getMin()*myForce->getMin());
  mySizeTest    = !periodic_boundary && (myCommunication == CommunicationOneSided || myCommunication == CommunicationOneSidedFast);
  if ((Real)(myNumberTotal/(myN*myM)) > myDensity){
    myMaxPerCell = (long)((Real)myNumberTotal/(Real)(myN*myM));
  }
  else{
    myMaxPerCell = (long)myDensity;
  }
  if(mySizeTest)
    myMaxPerCell *= 2;
  else
    myMaxPerCell= (long)((Real)myMaxPerCell*1.2);  
  myMaxPerCell = FastListImpl_best_new_count(myMaxPerCell, sizeof(Particle));

  strcpy(myOutFilename,world->getOutFilename());
  strcpy(myInFilename,world->getInFilename());
  output.open(myOutFilename, ios::out);
  myForce->setRCut(myRCut);

  // Some short cuts
  myDtSqrt = myDt*myDt/2.0;
  myDt2    = myDt*2.0;
  myDthalf = myDt/2.0;
  //myKsiT   = (Real)myNumber*KBOLZ*myTemp;
  myAlfa   = 0.0;
  myKsi    = 0.0;
  myTStep  = myTStop/(Real)myNStop;
  myLxN    = (Real)myN/myLx;
  myLyM    = (Real)myM/myLy;


  // Test if the # stops is too large
  while(myTStep+EPSILON < myDt){
    if(myId == master)
      cerr <<"Decreasing nstop.\n";
    myNStop /= 2;
    myTStep  = myTStop/(Real)myNStop;    
  }

  if (myN < 3 || myM < 3){
    if(myId == master)
      cerr << "Number of boxes to small!\n";
    myNStop = -1;
    delete world;
    return;
  }

  if (myN * myM < numProc*3){
    if(myId == master)
      cerr << "To many process per boxes!\n";
    myNStop       = -1;
    delete world;
    return;
  }

  if (numProc > max_number_proc){
    if(myId == master)
      cerr << "To many process!\n";
    myNStop       = -1;
    delete world;
    return;
  }

  // Init lists & arrays
  myCells           = new Cell[myN*myM];
  myOwnCells        = new List<CellIndex>(myN*myM/numProc+2);
  myOtherCells      = new List<CellIndex>;
  myOtherCellsUsed  = new List<CellIndex>;
  myMissingCells    = new List<CellIndex>(5);
  myTmpList         = new FastList<Particle>(0);
  myNeighbours      = new List<int>(5);
  myOverlap         = new List<int>(5);
  myRequestCell     = new List<CellIndex>(numProc);
  myRequestId       = new List<int>(numProc);
  myUpdateParticles = new FastList<Particle>[numProc];
  myIntArray        = new int[numProc];
  myIntArray2       = new int[numProc];
  myDumpTab         = new Real[myNStop+1];
  myBoundsX         = new Real[myN+1];
  myBoundsY         = new Real[myM+1];
  mySendList        = new List<CellIndex>;
  myRecvList        = new List<CellIndex>;
#ifdef HAVE_MPI
  myRequestArray    = new MPI_Request[numProc];
  myStatusArray     = new MPI_Status[numProc];
  myRequestList     = new FastList<MPI_Request>;
  myStatusList      = new FastList<MPI_Status>;
  for(i=0;i<numProc;i++)
    myRequestArray[i]=MPI_REQUEST_NULL;

  myNBuffer         = (int)(n1*sizeof(CellIndex)+n2*mymallocsize(Particle,myMaxPerCell));
  myBuffer          = malloc(myNBuffer);
  MPI_Buffer_attach(myBuffer,myNBuffer);
#endif
#ifdef HAVE_ONE
  TYPE_CELL           = MPI_DATATYPE_NULL;
  myTypeCount         = 4;
  mySharedCells       = new List<CellIndex>;
  myNumberSharedCells = new long[numProc];
  myShmem             = new void*[numProc];
  myWindow            = new MPI_Win[numProc];
  for(i=0;i<numProc;i++){      
    myNumberSharedCells[i] = 0;
    myShmem[i]             = NULL;
    myWindow[i]            = MPI_WIN_NULL;
  }
  myTmpLocalCells     = NULL;
#endif

  // Lookup table for the bounds
  myBoundsX[0]   = MINReal;
  myBoundsX[myN] = MAXReal;
  for(i=1;i<myN;i++)
    myBoundsX[i] = (Real)i/myLxN+myLeft;
  myBoundsY[0]   = MINReal;
  myBoundsY[myM] = MAXReal;
  for(i=1;i<myM;i++)
    myBoundsY[i] = (Real)i/myLyM+myBottom;

  // The dump time table 
  k = 0;
  l = (int)world->getDj();
  myDumpTab[k++] = 0.0;

  if (l < 1)
    l = 1;

  Real base = exp(log(myTStop)/(Real)l);
  Real last = 0.0;
  Real next = base;
  for(i=0;i<l;i++){
    long i0 = (myNStop*i)/l;
    long i1 = (myNStop*(i+1))/l;
    for(j=1;j<=i1-i0;j++){
      myDumpTab[k++] = last+(next-last)*(Real)j/(Real)(i1-i0);
    }
    last = next;
    next = next*base;
  }
   

  partitioning();
  neighbourhood();

  // The world knows best how to generate and add the particles 
  // to the universe. We have to adjust the number of particles (pull & lost)
  world->addParticles(this);
  myNumber = world->getNumber();

  myNumberTotal = myNumber;
#ifdef HAVE_MPI
  MPI_Allreduce(&myNumber,&myNumberTotal,1,MPI_LONG,MPI_SUM,MPI_COMM_WORLD);
#endif
  myKsiT   = (Real)myNumberTotal*KBOLZ*myTemp;

  // That's it for StartWorld
  delete world;

  // Prints
  if(myId == master){
    cerr << "Area: (" << myLeft <<","<< myBottom <<")-("<< myRight <<","<< myTop 
         <<"), (" << myLx <<","<< myLy <<")\n"
	 << "Number of boxes: " << myN <<" x "<< myM <<", ("<<myA<<","<<myB<<")\n"
	 << "Estimated average number of particles per box: "
         << (Real)myNumberTotal/(Real)(myN*myM)<<"\n"
	 << "Estimated max number ("<<myMaxPerCell<<") of particles per box: "
         <<myDensity<<"\n"      
	 << "Density: "<<(Real)myNumberTotal/((myLx)*(myLy))<<"\n"
	 << "Average number of neighbours per particle: "
         <<(M_PI*myRCut*myRCut*(Real)myNumberTotal)/(myLx*myLy)<<"\n"
	 << "Min R0: "<<myForce->getMin()<<"\n"
	 << "Timer overhead: estimated "<<overheadtime<<"[s], averaged "<<averagedoverheadtime<<"[s], minimal "<<minimaloverheadtime<<"[s].\n"
	 << "Dump times: ";
    for(i=0;i<myNStop;i++){
      cerr <<myDumpTab[i]<<", ";
      if(i > 0 && (i % 15) == 0)
	cerr <<"\n            ";
    }
    cerr <<myDumpTab[i]<<".\n";
#ifdef HAVE_MPI
    cerr <<"Bsend-Buffer size: "<<myNBuffer<<"\n";
#endif
  }

  for(j=0;j<numProc;j++){
#ifdef HAVE_MPI
    MPI_Barrier(MPI_COMM_WORLD);
#endif
    if(myId==j){
      cerr << "Proc "<< myId <<" overlap: ";
      for(i=0;i<myOverlap->count();i++)
	cerr << myOverlap->item(i)<<" ";
      cerr <<"\nProc "<< myId <<" neighbours: ";
      for(i=0;i<myNeighbours->count();i++)
	cerr << myNeighbours->item(i)<<" ";
      cerr <<"\nProc "<<myId<<" interactions: "<<myInteractions<<"\n";      
    }    
#ifdef HAVE_MPI
    MPI_Barrier(MPI_COMM_WORLD);
#endif
  }

  // Dump data
  dump();

  // The simulation starts .... zzzzZZZZZzzz
  long long localTimerCounter = timerCounter;
  myTimeTotal = mytime();

  // Forces
  force();

  localTimerCounter = timerCounter - localTimerCounter;
  myTimeSim  = mytime() - myTimeTotal-overheadtime*localTimerCounter;
}

Universe::~Universe(){

  if(getNStop()>0){
    myTimeTotal = mytime() - myTimeTotal-overheadtime;
    if(numProc <= 1){
      myTimeIdle = 0.0;
      myTimeDist = 0.0;
      myTimeCom  = 0.0;
      myTimeComPop  = 0.0;
      myTimeComPush  = 0.0;
      myTimeComGet  = 0.0;
    }
    myTimeCalc -= myTimeCom+myTimeUpdate+myTimeDist;
    myTimeCom  -= myTimeIdle;
    myTimeRest  = myTimeTotal-myTimeCalc-myTimeCom-myTimeIdle-myTimeUpdate-myTimeDist;

    long j = 0;
    for(long i=0;i<myOwnCells->count();i++)
      j += getCell(myOwnCells->item(i))->count();
    long l = j;
    long m = myCom;
    long p = myComBrute;
    double x = myTimeCalc;
    //    long long t = myParInter;
    long t = myParInter;
    
#ifdef HAVE_MPI
    MPI_Allreduce(&j,&l,1,MPI_LONG,MPI_SUM,MPI_COMM_WORLD);
    MPI_Allreduce(&myCom,&m,1,MPI_LONG,MPI_SUM,MPI_COMM_WORLD);
    MPI_Allreduce(&myComBrute,&p,1,MPI_LONG,MPI_SUM,MPI_COMM_WORLD);
    MPI_Allreduce(&myTimeCalc,&x,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);

    //    MPI_Allreduce(&myParInter,&t,1,MPI_LONG_LONG,MPI_SUM,MPI_COMM_WORLD);
    MPI_Allreduce(&myParInter,&t,1,MPI_LONG,MPI_SUM,MPI_COMM_WORLD);

#endif
    for(long k=0;k<numProc;k++){
      if(myId==k){
	cerr << "TimeTotal    :"<<myTimeTotal  <<".\n"
	     << "TimeSim      :"<<myTimeSim    <<" ("<<100*myTimeSim    /myTimeTotal<<"%).\n"
	     << "TimeCalc     :"<<myTimeCalc   <<" ("<<100*myTimeCalc   /myTimeTotal<<"%).\n"
	     << "TimeCom      :"<<myTimeCom    <<" ("<<100*myTimeCom    /myTimeTotal<<"%).\n"
	     << "TimeComPush  :"<<myTimeComPush<<" ("<<100*myTimeComPush/myTimeTotal<<"%).\n"
	     << "TimeComPop   :"<<myTimeComPop <<" ("<<100*myTimeComPop /myTimeTotal<<"%).\n"
	     << "TimeComGet   :"<<myTimeComGet <<" ("<<100*myTimeComGet /myTimeTotal<<"%).\n"
	     << "TimeIdle     :"<<myTimeIdle   <<" ("<<100*myTimeIdle   /myTimeTotal<<"%).\n"
	     << "TimeUpdate   :"<<myTimeUpdate <<" ("<<100*myTimeUpdate /myTimeTotal<<"%).\n"
	     << "TimeDist     :"<<myTimeDist   <<" ("<<100*myTimeDist   /myTimeTotal<<"%).\n"
	     << "TimeRest     :"<<myTimeRest   <<" ("<<100*myTimeRest   /myTimeTotal<<"%).\n"
	     << "CalcBalance  :"<<myTimeCalc*numProc*100/x-100<<"% ("<<x<<").\n"
	     << "Com          :"<<myCom/myDtCount<<".\n";
	if(p>0)
	  cerr << "ComTot       :"<<m/myDtCount<<" ("<<100*m/p<<"%).\n";
	cerr << "ComBrute     :"<<myComBrute/myDtCount<<".\n"
	     << "ComBruteTot  :"<<p/myDtCount<<".\n"
	     << "Interactions :"<<myComInter/myDtCount<<".\n"
	     << "Distribute   :"<<myDistribute<<".\n"
	     << "ParInter     :"<<2.0*(double)t/(Real)((myDtCount+1)*l)<<".\n"
	     << "Particles    :"<<j<<".\n"
	     << "Total Par    :"<<l<<".\n"  
	     << "mytime       :"<<timerCounter<<","<<timerCounter*overheadtime<<"[s].\n";    
      }
#ifdef HAVE_MPI
      MPI_Barrier(MPI_COMM_WORLD);
#endif
    }    
  }

  output.close();

  delete myForce;
  delete myOwnCells;
  delete [] myCells;
  delete myOtherCells;
  delete myOtherCellsUsed;
  delete myMissingCells;
  delete myTmpList;
  delete myOverlap;
  delete myNeighbours;
  delete myRequestCell;
  delete myRequestId;
  delete [] myUpdateParticles;
  delete [] myIntArray;
  delete [] myIntArray2;
  delete mySendList;
  delete myRecvList;
  delete [] myDumpTab;
  delete [] myBoundsX;
  delete [] myBoundsY;
#ifdef HAVE_MPI
  delete [] myRequestArray;
  delete [] myStatusArray;
  delete myRequestList;
  delete myStatusList;
  int size = 0;
  MPI_Buffer_detach(myBuffer,&size);
  free(myBuffer);
#endif
#ifdef HAVE_ONE
  delete mySharedCells;
  for(long k=0;k<numProc;k++){      
    myfreeShm(myShmem[k]);
    if(myWindow[k] != MPI_WIN_NULL)
      MPI_Win_free(&myWindow[k]);
  }
  delete [] myNumberSharedCells;
  delete [] myShmem;
  delete [] myWindow;
  if(TYPE_CELL != MPI_DATATYPE_NULL)
    MPI_Type_free(&TYPE_CELL);
  myfree(myTmpLocalCells);
#endif

}

void Universe::partitioning(){

  // The weight of each cell
  long i,j,k,l;
  myWeightSum = 0;
  for(j=0;j<myM;j++){
    for(i=0;i<myN;i++){
      getCell(i,j)->setIndex(CellIndex(i,j));
      if(!periodic_boundary){
	if(j==myM-1 && i==myN-1)
	  getCell(i,j)->setWeight(1);
	else if(j==myM-1)
	  getCell(i,j)->setWeight(3);
	else if (i == 0)
	  getCell(i,j)->setWeight(7);
	else if (i == myN-1)
	  getCell(i,j)->setWeight(5);
	else
	  getCell(i,j)->setWeight(9);
      }
      else if(!periodic_boundary_y){
	if(j==myM-1)
	  getCell(i,j)->setWeight(3);
	else
	  getCell(i,j)->setWeight(9);
      }
      else if(!periodic_boundary_x){
	if(i==myN-1)
	  getCell(i,j)->setWeight(5);
	else if (i == 0)
	  getCell(i,j)->setWeight(7);
	else
	  getCell(i,j)->setWeight(9);
      }
      else{
	getCell(i,j)->setWeight(1);
      }
      myWeightSum += getCell(i,j)->getWeight();     
    }
  }

  // Partition
  int* id = new int[myN*myM];
  for(i=0;i<myN*myM;i++)
    id[i] = 0;

  if(numProc > 1){
    // Stripes, default
    if((myCommunication == CommunicationSimple || myCommunication == CommunicationSequential) && periodic_boundary_x){
      Real w = (Real)myM;
      if (!periodic_boundary_y)
	w -= 2.0/3.0;
      k = 0;
      for(j=0;j<myM;j++){
	l = (int)((Real)(numProc*j)/w);
	if(l >= numProc)
	  l = numProc - 1;
	for(i=0;i<myN;i++){
	  id[k] = (int)l;
	  k++;
	}
      }
    }
    else {
      k = 0;
      l = 0;
      for(j=0;j<myM;j++){
	for(i=0;i<myN;i++){
	  id[k] = (int)(numProc*l/(myWeightSum)); 
	  k++;
	  l += getCell(i,j)->getWeight();
	}
      }
    }

    // Checker
    if(myPartition == PartitionChecker){
      k = 0;
      l = 0;
      for(j=0;j<myM;j++){
	for(i=0;i<myN;i++){
	  id[k] = ((int)k)%numProc;
	  k++;
	  l += getCell(i,j)->getWeight();
	}
      }  
    }
    // Metis
    else if(myPartition == PartitionMetis){
#ifdef HAVE_METIS
      if(myId == master){
	idxtype* metisid = new idxtype[myN*myM];
	for(i=0;i<myN*myM;i++)
	  metisid[i] = 0;
      
	long p,m;
	int n,wgtflag,numflag,nparts,edgecut;
	n = (int)(myN*myM);
	int options[] = {0,0,0,0,0};
	numflag = 0;
	nparts = numProc;
	
	idxtype* xadj = new idxtype[n+1];
	idxtype* adjncy = new idxtype[n*4];
	idxtype* vwgt = NULL;
	
	if(!periodic_boundary){
	  wgtflag = 2;
	  vwgt = new idxtype[n];
	}
	else {
	  wgtflag = 0;
	}
	
	k = 0;
	m = 0;
	xadj[0] = 0;
	for(j=0;j<myM;j++){
	  for(i=0;i<myN;i++){
	    xadj[k+1] = xadj[k];
	    if(vwgt != NULL){
	      vwgt[k]= getCell(i,j)->getWeight();
	      for(l=-1;l<=1;l++)
		for(p=-1;p<=1;p++)
		  if(!(l == 0 && p == 0) && l*p == 0 && 
		     i+l<myN && i+l>=0 && j+p<myM && j+p>=0){
		    adjncy[m++] = (int)(i+l+myN*(j+p));
		    xadj[k+1]++;
		  }	  
	    }
	    else {
	      for(l=-1;l<=1;l++)
		for(p=-1;p<=1;p++)
		  if((l != 0 || p != 0) && l*p == 0){
		    adjncy[m++] = (int)((i+l+myN)%myN+myN*((j+p+myM)%myM));
		    xadj[k+1]++;
		  }	  
	    }
	    k++;
	  }
	}
	METIS_PartGraphRecursive(&n, xadj,adjncy,vwgt,NULL,&wgtflag,&numflag,&nparts,options,&edgecut,metisid);
	
	for(i=0;i<myN*myM;i++)
	  id[i] = metisid[i];
	
	delete [] xadj;
	delete [] adjncy;
	delete [] vwgt;
	delete [] metisid;
      }
#ifdef HAVE_MPI
      MPI_Bcast(id,(int)(myN*myM), MPI_INT, master,MPI_COMM_WORLD);
#endif
#endif
    }
    else if(myPartition == PartitionRectangle){
      long n,m;
      
      n = 1;
      m = numProc;
      for(i=2;i<=numProc;i++){
	if((numProc % i) == 0){
	  //if(myId == master)
	  //  cerr <<i<<"*"<<numProc / i<<"="<<numProc<<":("<<n<<","<<m<<") "
	  // << myN*m+2*myM*n<<","<<n*m <<","<<myN*numProc/i+2*myM*i<<","
	  // <<i*(numProc/i) <<"\n";
	  if((myN*m+2*myM*n-n*m)>(myN*numProc/i+2*myM*i-i*(numProc/i))){
	    n = i;
	    m = numProc/i;
	  }
	}
      }
			       
      k = 0;
      for(long j=0;j<myM;j++){
	for(long i=0;i<myN;i++){
	  id[k] = (int)((numProc*j/n)/myM)*(int)n+(int)((numProc*i/m)/myN);
	  k++;
	}
      }
    }
  }

  // Set owner and actual proc id of each cell
  k = 0;
  for(j=0;j<myM;j++){
    for(i=0;i<myN;i++){
      getCell(i,j)->setOwnerId(id[k]);
      getCell(i,j)->setActualId(id[k]);
      k++;
    }
  }
  delete [] id;

  // Find all cells which belonging to myId proc
  myOwnCells->removeAll();
  for(j=0;j<myM;j++)
    for(i=0;i<myN;i++)
      if(myId == getOwnerId(i,j))
	myOwnCells->append(CellIndex(i,j));

#ifdef HAVE_MPI
  // Print the partition
  k = (myN + myM) / 200 + 1;
  if(myId == master && numProc > 1){
    for(j=myM-1;j>=0;j--){
      for(i=0;i<myN;i++){
	if(i % k == 0 && j % k == 0){
	  if(myN >40)
	    cerr << id_char[getOwnerId(i,j)] ;
	  else
	    cerr << id_char[getActualId(i,j)]
		 << id_char[getOwnerId(i,j)];
	}
      }
      if(j % k == 0)
	cerr <<"\n";
    }
  }
  MPI_Barrier(MPI_COMM_WORLD);
#endif
}

void Universe::neighbourhood(){

  long i,j,k;
  myInteractions = 0;
  myOverlap->removeAll();
  myNeighbours->removeAll();
  mySendList->removeAll();
  myRecvList->removeAll();
  myOtherCells->removeAll();
  myOtherCellsUsed->removeAll();
  myMissingCells->removeAll();
  for(j=0;j<myM;j++){
    for(i=0;i<myN;i++){
      Cell* cell = getCell(i,j);
      cell->setPath(NULL);  
      cell->usedByReset();
      cell->reduce();
    }
  }

  // Define the path of interactions
  // Compute the number of interaction between cells for each proc
  for(k=0;k<myOwnCells->count();k++){
    CellIndex index = myOwnCells->item(k);
    i = index.i();
    j = index.j();
    List<CellIndex>* path = new List<CellIndex>(5);
    path->append(index);
    if((periodic_boundary_x || i>0) && (periodic_boundary_y || j+1 <myM))
      path->append(CellIndex((i-1+myN)%myN,(j+1)%myM));
    if(periodic_boundary_y || j+1 <myM)
      path->append(CellIndex( i       ,(j+1)%myM));
    if((periodic_boundary_x || i+1 <myN) && (periodic_boundary_y || j+1 <myM))
      path->append(CellIndex((i+1)%myN,(j+1)%myM));
    if(periodic_boundary_x || i+1 <myN)
      path->append(CellIndex((i+1)%myN, j       ));
    getCell(i,j)->setPath(path);  
    myInteractions +=path->count();
    // The overlap
    for(i=0;i<path->count();i++){
      j = getCell(path->item(i))->getOwnerId();
      if(j != myId){
	myOverlap->appendUnique((int)j);
      }
    }
  }


  // Neighbours which need my cells
  for(j=0;j<myM;j++){
    for(i=0;i<myN;i++){
      int id1 = getOwnerId(i,j);

      if((periodic_boundary_x || i>0) && (periodic_boundary_y || j+1 <myM)){
	Cell* cell = getCell((i-1+myN)%myN,(j+1)%myM);
	if(id1 != myId && cell->getOwnerId() == myId)
	  myNeighbours->appendUnique(id1);
	cell->usedByAdd(id1);
      }
      if(periodic_boundary_y || (j+1 <myM)){
	Cell* cell = getCell(i       ,(j+1)%myM);
	if(id1 != myId && cell->getOwnerId() == myId)
	  myNeighbours->appendUnique(id1);
	cell->usedByAdd(id1);
      }
      if((periodic_boundary_x || i+1 <myN) && (periodic_boundary_y || j+1 <myM)){
	Cell* cell = getCell((i+1)%myN,(j+1)%myM);
	if(id1 != myId && cell->getOwnerId() == myId)
	  myNeighbours->appendUnique(id1);
	cell->usedByAdd(id1);
      }
      if(periodic_boundary_x || (i+1 <myN)){
	Cell* cell = getCell((i+1)%myN, j       );
	if(id1 != myId && cell->getOwnerId() == myId)
	  myNeighbours->appendUnique(id1);
	cell->usedByAdd(id1);
      }      
    }
  }

  // Cells used, but hold by other procs
  for(j=0;j<myM;j++){
    for(i=0;i<myN;i++){
      Cell* cell = getCell(i,j);
      for(k=0;k<max_neighbours;k++){
	if(cell->usedBy(k)<0)
	  break;
	if(myId == cell->usedBy(k)){
	  myOtherCellsUsed->append(cell->getIndex());
	  break;
	}
      }
    }
  }


  // For the simple/ sequential (MPI) 
  for(j=0;j<myM;j++){
    for(i=0;i<myN;i++){
      if (getCell(i,j)->own() && getCell(i,j)->usedByAny()){
	mySendList->append(CellIndex(i,j));
      }
      else if (myId == getCell(i,j)->usedBy(0)){
	myRecvList->append(CellIndex(i,j));
      }
    }
  }

#ifdef HAVE_ONE 
  myTypeCount = 4;
  for(i=0;i<numProc;i++){
    
    if(myWindow[i] != MPI_WIN_NULL)
      MPI_Win_free(&myWindow[i]);
    myfreeShm(myShmem[i]);
    myShmem[i] = NULL;
    myNumberSharedCells[i] = 0;
  }
  if(TYPE_CELL != MPI_DATATYPE_NULL)
    MPI_Type_free(&TYPE_CELL);
  mySharedCells->removeAll();


  if(myCommunication == CommunicationOneSided ||
     myCommunication == CommunicationOneSidedFast){
    // mySharedCells, List of owend, shared cells
    for(j=0;j<myM;j++){
      for(i=0;i<myN;i++){
	if(getCell(i,j)->usedByAny()){
	  myNumberSharedCells[getCell(i,j)->getOwnerId()]++;
	  if(getCell(i,j)->own()){
	    mySharedCells->append(getCell(i,j)->getIndex());
	    //cerr << myId <<" : "<<(CellIndex&)getCell(i,j)->getIndex()<<"\n";
	  }
	}
      }
    }  
    
    // Allocating memory for the shared cells
    // Creating window for RMA
    if(myTmpLocalCells != NULL)
      myfree(myTmpLocalCells);
    myTmpLocalCells = mymallocArray(Particle,myMaxPerCell,myNumberSharedCells[myId]);

    for(k=0;k<numProc;k++){          
      
      long count = 0;
      
      for(i=0;i<myOverlap->count();i++){
	if(myOverlap->item(i) == k ){
	  count = myNumberSharedCells[k];
	  break;
	}
      }

      if(k == myId)
	count = myNumberSharedCells[k];
	
      if (count > 0)
	myShmem[k]  = mymallocShmArray(Particle,myMaxPerCell,count);      
      else
	myShmem[k]  = mymallocShm(Particle,0);

      if(count > 0)
	cerr << myId <<","<<k<<" : mymallocShmArray ("
	     <<myMaxPerCell<<","<<count<<") = "<<mymallocsizeArray(Particle,myMaxPerCell,count)<<"\n";

      MPI_Win_create(myShmem[k],mymallocsizeArray(Particle,myMaxPerCell,count),1,MPI_INFO_NULL,MPI_COMM_WORLD,&myWindow[k]);
      MPI_Win_fence(0,myWindow[k]);

      char* ptr = (char*)myShmem[k];
      for(j=0;j<myM;j++){
	for(i=0;i<myN;i++){
	  Cell* cell = getCell(i,j);
	  //MPI_Barrier(MPI_COMM_WORLD);
	  if(cell->getOwnerId() == k && cell->usedByAny() && count > 0){
	    cell->setShmPtr((void*)ptr);
	    ptr +=mymallocsize(Particle,myMaxPerCell);
	  }
	  //MPI_Barrier(MPI_COMM_WORLD);
	}
      }    
    }

    // Make sure that the shared cells have enough space
    for(j=0;j<myM;j++){
      for(i=0;i<myN;i++){
	Cell* cell = getCell(i,j);
	// Shared, owned
	if(cell->own() && cell->usedByAny()){
	  cell->atLeastSize(myMaxPerCell-1);
	}
	// Shared, hold by other proc
	else {
	  for(k=0;k<max_neighbours;k++){
	    if(cell->usedBy(k) == myId){
	      cell->atLeastSize(myMaxPerCell-1);
	      break;
	    }
	  }
	} 
      }
    }
    
    
    build_derived_type_Cell(&TYPE_CELL, myMaxPerCell);
    
  }
#endif

}

ostream& operator<<(ostream& OS, Universe & universe){
  OS << "Universe\n"
     << " Step : "<<universe.myCount<<"\n";

  long k;
  for(k=0;k<universe.myOwnCells->count();k++){
    CellIndex index = universe.myOwnCells->item(k);
    OS << "Cell "<<index <<": "<< universe.getCell(index)->count()<<" ("<<universe.getCell(index)->getParticles()->size()<<")\n";

  }  
  return OS;
}

void Universe::next(){

  double sim = mytime();
  //#ifdef HAVE_MPI
  //MPI_Barrier(MPI_COMM_WORLD);
  //  myTimeIdle += mytime()-sim;
  //#endif
  Real t2 = (Real)(myCount+1)*myTStep;
  long count = myOwnCells->count();

  // Compute the interactions/force
  for(;((Real)myDtCount*myDt-EPSILON)<t2;){
    long long localTimerCounter = timerCounter;
    double calc = mytime();
    myTime = (Real)myDtCount*myDt;
#ifdef HAVE_PULL
    bool newStep = true;
#endif
    long j,k,l;
    long double ekin = 0;
    long double fa = 0;
    //#pragma omp parallel for private (j,k,l) reduction(+:ekin,fa) reduction(*:newStep)
    for(j=0;j<count;j++){
      Cell* cell     = getCell(myOwnCells->item(j));
      long count2    = cell->count();
      Particle* pars = cell->getItems();
      for(k=0;k<count2;k++){
 	Particle& par = pars[k];
 	//Particle par = pars[k];
#ifdef HAVE_PULL
	if(par.tag < 3){
#endif
	  //par.x += par.vx*myDt+myDtSqrt*(par.fxa+myForce->fx-myKsi*par.vx)/par.mass;
	  //par.y += par.vy*myDt+myDtSqrt*(par.fya+myForce->fy-myKsi*par.vy)/par.mass;
      	  Real x = par.x + par.vx*myDt+myDtSqrt*(par.fxa+myForce->fx-myKsi*par.vx)/par.mass;
  	  Real y = par.y + par.vy*myDt+myDtSqrt*(par.fya+myForce->fy-myKsi*par.vy)/par.mass;
	  if(periodic_boundary_x){
	    if(x<myLeft)
	      x += myLx;
	    else if (x>=myRight)
	      x -= myLx;
	  }
	  if(periodic_boundary_y){
	    if(y<myBottom)
	      y += myLy;
	    else if (y>=myTop)
	      y -= myLy;
	  }
	  par.x = x;
	  par.y = y;
	  par.vx += myDthalf*(par.fxa+myForce->fx-myKsi*par.vx)/par.mass;
	  par.vy += myDthalf*(par.fya+myForce->fy-myKsi*par.vy)/par.mass;
#ifdef HAVE_PULL
	}
	else {
	  if(myPullStart <= myTime && myTime < myPullStop){
	    par.y += par.vy*myDt;
	    if(newStep){
	      myL1 += par.vy*2*myDt*sign(par.y);
	      newStep = false;
	    }
	  }
	}
#endif
	par.fxa  = 0.0;
	par.fya  = 0.0;
	par.epot = 0.0;
	//pars[k] = par;
      }
    }

    update();

    force();

    ekin = 0;
    fa = 0;
    l = 0;

    for(j=0;j<count;j++){
      Cell* cell = getCell(myOwnCells->item(j));
      long count2 = cell->count(); 
      Particle* pars = cell->getItems();
      for(k=0;k<count2;k++){
	Particle& par = pars[k];
#ifdef HAVE_PULL
	if(par.tag < 3){
#endif
	  par.vx += myDthalf*(par.fxa+myForce->fx-myKsi*par.vx)/par.mass;
	  par.vy += myDthalf*(par.fya+myForce->fy-myKsi*par.vy)/par.mass;
#ifdef HAVE_PULL
	}
	else{
	  fa += sign(par.y)*par.fya;
	  l++;
	}
#endif
	ekin += 0.5*par.mass*(par.vx*par.vx+par.vy*par.vy);
      }
    }

    long double epot = myEpot;
#ifdef HAVE_MPI
    if (numProc > 1) {
      // Collect values from all proc's
      double com = mytime();
      double sum[]   = {0.0,  0.0, 0.0,  0.0};
      double local[] = {ekin, fa,  epot, (Real)l};
      MPI_Allreduce(local,sum,4,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
      ekin = sum[0];
      fa   = sum[1];
      epot = sum[2];
      l    = (long)sum[3];      
      myTimeCom += mytime()-com-overheadtime;
    }
#endif
    myKsi = (ekin-myKsiT)*myAlfa;
    if(l>0)
      fa /= l;

    myEkin = ekin;
    myFa   = fa;
    myEpot = 2*epot;

    localTimerCounter = timerCounter - localTimerCounter;
    myTimeCalc += mytime() - calc - overheadtime*localTimerCounter;

    // Update the cells
    update();

    myDtCount++;
    myTime =  (Real)myDtCount*myDt;

    // Dump data
    dump();

  }

  myCount++;
  myTimeSim += mytime()-sim-overheadtime;

}

void Universe::update(){
  double update = mytime();

  int p;
  for(p=0;p<numProc;p++)
    myUpdateParticles[p].removeAll();

  // Update the cells
  long count = myOwnCells->count();
  long n = 0;
  long k;
  for(long i=0;i<count;i++){
    Cell* cell = getCell(myOwnCells->item(i));
    long count2 = cell->count(); 
    Particle* par = cell->getItems();
    long i0 = (myOwnCells->item(i)).i();
    long j0 = (myOwnCells->item(i)).j();
    Real x0 = myBoundsX[i0];
    Real x1 = myBoundsX[i0+1];
    Real y0 = myBoundsY[j0];
    Real y1 = myBoundsY[j0+1];
    for(k=count2-1;k>=0;k--){
      if (par[k].x < x0 || x1 <= par[k].x || par[k].y < y0 || y1 <= par[k].y ){
        CellIndex indexNew(getIndex(par[k]));
	int id = getOwnerId(indexNew);
	if(id != myId)
	  myUpdateParticles[id].add(par[k]);
	else{
	  getCell(indexNew)->add(par[k]);
	  if(mySizeTest && getCell(indexNew)->size() > myMaxPerCell)
	    cerr <<"To many particles\n";
	  n++;
	}
	cell->remove(k);
      }
    }
  }

  double distribute = mytime();
  myTimeUpdate += distribute-update-overheadtime;
#ifdef HAVE_MPI
  // Redistibute the particles, which belong other proc's
  if (numProc > 1){
    // Send the particles
    for(p=0;p<numProc;p++){
      if (myUpdateParticles[p].count() > 0 && p!=myId){
	MPI_Isend(myUpdateParticles[p].getItems(),(int)myUpdateParticles[p].count(),TYPE_PARTICLE,p,UPDATE_CELLS,MPI_COMM_WORLD,&myRequestArray[p]);
	myDistribute +=myUpdateParticles[p].count();
	myIntArray[p]=1;
	myNumber -= myUpdateParticles[p].count();
      }
      else{
	myIntArray[p]=0;
	myRequestArray[p]=MPI_REQUEST_NULL;
      }
    }
    
    // Compute the number of receives
    MPI_Allreduce(myIntArray,myIntArray2,numProc,MPI_INT,MPI_SUM,MPI_COMM_WORLD);

    // Recieve the particles and put them into the right cell
    for(p=0;p<myIntArray2[myId];p++){
      MPI_Status status;
      MPI_Probe(MPI_ANY_SOURCE,UPDATE_CELLS,MPI_COMM_WORLD, &status);
      int id = status.MPI_SOURCE;
      int count;
      MPI_Get_count(&status,TYPE_PARTICLE,&count);
      MPI_Recv(myUpdateParticles[myId].setItems(count),count,TYPE_PARTICLE,id,UPDATE_CELLS,MPI_COMM_WORLD, &status);
      for(k=0;k<myUpdateParticles[myId].count();k++){
	addParticle(Particle(myUpdateParticles[myId].item(k)));
	if(mySizeTest && getCell(myUpdateParticles[myId].item(k))->size() > myMaxPerCell)
	  cerr <<"To many particles\n";

      }
      myNumber += count;
    }
    MPI_Waitall(numProc,myRequestArray,myStatusArray);
    myTimeDist += mytime()-distribute-overheadtime;
  }
#endif
}

void Universe::force(){
  //
  // We assume that fxa,fya and epot of each particle is zero/ cleared.
  //

  long count = myOwnCells->count();
  myNeighboursCount = myNeighbours->count();
  myEpot = 0;
  
  if(myCommunication == CommunicationSequential)
    sendRecvCellSeq(true);
  else if(myCommunication == CommunicationOneSided)
    pushSharedCells();
  else if(myCommunication == CommunicationOneSidedFast)
    pushSharedCellsFast();

  for(long i=0;i<count;i++){
    List<CellIndex>* path = getCell(myOwnCells->item(i))->getPath();
    Cell* src   = getCell(myOwnCells->item(i)); 
    long count2 = path->count();
    myComInter += count2-1;
    for(long j=0; j<count2;j++){

      Cell* dest = getCell(path->item(j));

      if(myCommunication == CommunicationDistributed)
	sendRecvCell(src,dest);
      else if(myCommunication == CommunicationOneSided)
	getRemoteCell(dest);
      else if(myCommunication == CommunicationSimple){
	if(j == 0)
	  sendRecvCellSimple(i);
      }

      // We take care of the boundaries conditions
      Real bx = 0.0;
      Real by = 0.0;
      if(periodic_boundary_x){
	long p = src->getIndex().i() - dest->getIndex().i();
	if (p < -1)
	  bx = -myLx;
	else if (p > 1)
	  bx = myLx;
      }
      if(periodic_boundary_y){
	long p = src->getIndex().j() - dest->getIndex().j();
	if (p < -1)
	  by = -myLy;
	else if (p > 1)
	  by = myLy;
      }
      computeInteraction(src,dest,bx,by);



      if(!dest->own())
	myComBrute+=2;
      
    }
  }

  if(myCommunication == CommunicationDistributed)
    sendRecvCell(NULL,NULL);
  else if(myCommunication == CommunicationSimple)
    sendRecvCellSimple(-1);
  else if(myCommunication == CommunicationSequential)
    sendRecvCellSeq(false);
  else if(myCommunication == CommunicationOneSided)
    popSharedCells();
  else if(myCommunication == CommunicationOneSidedFast)
    popSharedCellsFast();

}

void Universe::computeInteraction(Cell* src, Cell* dest, Real bx, Real by){

  // Interaction between particles in the same cell
  if(src == dest){
    long srcCount = src->count();
    Particle*  srcPar = src->getItems();

    for(long i=0;i<srcCount;i++){
      for(long j=i+1;j<srcCount;j++){
	if (myForce->calc(srcPar[i],srcPar[j],myEpot,bx,by))
	  myParInter++;
      }
    }
  }
  // Interaction between to diffrent cells
  else{
    long destCount = dest->count();
    long srcCount = src->count();
    Particle* destPar = dest->getItems();
    Particle*  srcPar = src->getItems();
    
    for(long i=0;i<srcCount;i++){
      for(long j=0;j<destCount;j++){
	if(myForce->calc(srcPar[i],destPar[j],myEpot,bx,by))
	  myParInter++;
      }
    }
  }
}


#ifdef HAVE_MPI
void Universe::sendRecvCell(Cell* src, Cell* dest){
  if (numProc < 2) return;

  
  double com = mytime();
  MPI_Status status;
  long i,j;
  Cell* cell = NULL;
  int count,size,id;
  CellIndex index;
  int tag;
  bool already;
  long backReq = 0;

  //debug <<"sendRecvCell in\n";
  //if(src != NULL)
  //    debug <<"src "<<(CellIndex&)src->getIndex()<<":"<<src->getOwnerId()<<":"<<src->getActualId()<<"\n";

  //if(dest != NULL)
  //    debug <<"dest "<<(CellIndex&)dest->getIndex()<<":"<<dest->getOwnerId()<<":"<<dest->getActualId()<<"\n";
  //debug <<flush;

  // Trying to serve the old requests
  for(i=myRequestId->count()-1;i>=0;i--){
    index   = myRequestCell->item(i);
    cell    = getCell(index);
    id      = myRequestId->item(i);
    already = (cell->getActualId() == id);

    // Do not send if we need them, the cell is not back yet, maybe next time
    if(src == cell || (dest == cell && cell->own()) || !cell->have())
      continue;

    myRequestCell->remove(i);
    myRequestId->remove(i);

    // Cell was already sent
    if(already)
      continue;

    if(id == cell->getOwnerId()){
      // We send back a missing cell, which will be send to the right proc
      MPI_Bsend(&index,1,TYPE_CELLINDEX,id,SEND_BACK_CELL,MPI_COMM_WORLD);
      MPI_Bsend(cell->getItems(),(int)cell->count(),TYPE_PARTICLE,id,SEND_BACK_CELL,MPI_COMM_WORLD);
    }
    else{
      // We send the cell of an old request
      MPI_Bsend(cell->getItems(),(int)cell->count(),TYPE_PARTICLE,id,SEND_DEST_CELL,MPI_COMM_WORLD);
    }

    cell->setActualId(id);
    if (cell->own())
      // Cell goes away
      myMissingCells->appendUnique(index);
    else
      // Cell goes back home
      myOtherCells->removeItem(index);
  }
  //debug <<"sendRecvCell A\n"<<flush;

  if (dest == NULL || src == NULL){
    // We finished and send back all cells which we do not own
    // We tell all proc we possible have asked for cells
    myRequestList->removeAll();
    for(i=0;i<myOverlap->count();i++){
      backReq++;
      id = myOverlap->item(i);
      count = 0;
      for(j=0;j<myOtherCells->count();j++)
	if(getOwnerId(myOtherCells->item(j)) == id && getActualId(myOtherCells->item(j)) == myId)
	  count++;

      MPI_Bsend(&count,1,MPI_INT,id,SEND_BACK_ALL_CELLS,MPI_COMM_WORLD);      
      for(j=myOtherCells->count()-1;j>=0;j--){
	cell = getCell(myOtherCells->item(j));
	index = cell->getIndex();
	if (cell->getOwnerId() == id){
	  if(cell->have()){
	    MPI_Request request;
	    MPI_Bsend(&index,1,TYPE_CELLINDEX,id,SEND_INDEX,MPI_COMM_WORLD);
	    MPI_Bsend(cell->getItems(),(int)cell->count(),TYPE_PARTICLE,id,SEND_OTHER_CELL,MPI_COMM_WORLD);
	    myRequestList->add(request);
	  }

	  myOtherCells->remove(j);
	  cell->setActualId(id);
	}
      }
    }
    for(i=myRequestId->count()-1;i>=0;i--){
      if(getCell(myRequestCell->item(i))->getActualId() == myRequestId->item(i)){
	myRequestCell->remove(i);
	myRequestId->remove(i);
	continue;
      }
    } 
  }
  
  //debug <<"sendRecvCell B\n"<<flush;

  bool flag    = true;
  bool reqSrc  = false;
  bool reqDest = false;
  int test = 0;
  
  while(flag){
    if (!reqSrc && src != NULL && !src->have()){
      // We need the src cell and ask for it
      MPI_Bsend((void*)&src->getIndex(),1,TYPE_CELLINDEX,src->getActualId(),REQUEST_SRC_CELL,MPI_COMM_WORLD);
      reqSrc = true;
    }
    if (!reqDest && dest != NULL && !dest->have() && dest != src){
      // We need the dest cell and ask for it
      MPI_Bsend((void*)&dest->getIndex(),1,TYPE_CELLINDEX,dest->getActualId(),REQUEST_DEST_CELL,MPI_COMM_WORLD);
      reqDest = true;	
    }
    if((src == NULL || dest == NULL) && myRequestCell->count()>0){
      for(i=myRequestId->count()-1;i>=0;i--){
	index   = myRequestCell->item(i);
	cell    = getCell(index);
	id      = myRequestId->item(i);
	already = (cell->getActualId() == id);
	
	// Cell was already sent
	if(already){
	  myRequestCell->remove(i);
	  myRequestId->remove(i);
	  continue;
	}

	// The cell is not back yet, maybe next time
	if(!cell->have()){
	  MPI_Bsend((void*)&cell->getIndex(),1,TYPE_CELLINDEX,cell->getActualId(),REQUEST_BACK_CELL,MPI_COMM_WORLD);
	  //debug <<"Proc "<<myId<<" retrying a back inner "<<(CellIndex&)cell->getIndex()<<".\n" <<flush;
	}
      }
    }
    
    MPI_Iprobe(MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD, &test, &status);
    
    // We wait or only check if we got a something
    if ((test == 0) && (reqSrc || reqDest || (backReq>0)|| ((dest == NULL || src == NULL) && myNeighboursCount >0))){
      
      // To detect deadlocks
      double idle = mytime();
      bool hanging = false;
      bool first   = true;
      bool wait    = true;

      while(wait){
	if(mytime()-idle > timeout_deadlock && first){
	  first = false;
	  if(src == NULL || dest == NULL){
	    cerr << "Proc "<< myId <<" waiting completion ("<<myRequestId->count()<<":";
	    for(i=0;i<myRequestCell->count();i++)
	      cerr <<(CellIndex&)myRequestCell->item(i)<<","<<myRequestId->item(i)<<","<< getActualId(myRequestCell->item(i))<<","<< getOwnerId(myRequestCell->item(i));
	    cerr <<").\n";
	  }
	  else {
	    cerr << "Proc "<< myId <<" waiting "<<(CellIndex&)src->getIndex()<<(CellIndex&)dest->getIndex();
	    if(reqSrc){
	      cerr <<", src "<<(CellIndex&)src->getIndex()<<" from "<<src->getActualId()<<" ("<<src->getOwnerId()<<")";
	      reqSrc = false;
	      wait = false;
	    }
	    if(reqDest){
	      cerr <<", dest "<<(CellIndex&)dest->getIndex()<<" from "<<dest->getActualId()<<" ("<<dest->getOwnerId()<<")";
	      reqDest = false;
	      wait = false;
	    }
	    cerr <<"("<<myRequestId->count()<<":";

	    for(i=0;i<myRequestCell->count();i++)
	      cerr <<(CellIndex&)myRequestCell->item(i)<<","<<myRequestId->item(i)<<","<< getActualId(myRequestCell->item(i))<<","<< getOwnerId(myRequestCell->item(i));
	    cerr <<").\n";
	  }
	}

	MPI_Iprobe(MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD, &test, &status);
	if(test == 0)
	  hanging = true;
	wait &= (test == 0);
      }

      if(hanging){
	idle = mytime()-idle;
	myTimeIdle += idle;
	if(idle > timeout_deadlock)
	  cerr << "Proc "<< myId <<" ok ("<<idle<<").\n";
      }
    }

    flag = (test != 0);

    if(flag){
      switch (status.MPI_TAG) {
      case NOT_YET:
	//debug << "NOT_YET\n"<<flush; 
	// The dest cell is at the moment not available, only from a dest cell
	MPI_Recv(NULL,0,MPI_CHAR,status.MPI_SOURCE,NOT_YET,MPI_COMM_WORLD,&status);
	//debug << "NOT_YET done\n"<<flush; 
	break;
	
      case REQUEST_SRC_CELL:
      case REQUEST_DEST_CELL:
	// There was a request for a cell and we send it to the proc if possible
	id  = status.MPI_SOURCE;
	tag = status.MPI_TAG;
	//debug << "REQUEST_*_CELL "<<id <<","<<tag<<"\n"<<flush; 
	MPI_Recv(&index,1,TYPE_CELLINDEX,id,tag,MPI_COMM_WORLD,&status);
	cell    = getCell(index);
	already = (cell->getActualId() == id);

	if(already){
	  //debug << "REQUEST_*_CELL done\n"<<flush; 
	  break;
	}
	//debug << "REQUEST_*_CELL A\n"<<flush; 
	if(src == cell || (dest == cell && cell->own())){
	  // We do not give away the cells which are needed for actual interaction
	  // We will send it later
	  MPI_Bsend(NULL,0,MPI_CHAR,id,NOT_YET,MPI_COMM_WORLD);
	  bool unique = true;
	  for(i=0;i<myRequestCell->count();i++)
	    if(myRequestCell->item(i) == index && myRequestId->item(i) == id){
	      unique = false;
	      //debug << "REQUEST_*_CELL done\n"<<flush; 
	      break;
	    }
	  if(unique){
	    myRequestId->prepend(id);
	    myRequestCell->prepend(index);
	  }
	  //debug << "REQUEST_*_CELL done\n"<<flush; 
	  break;
	}
	//debug << "REQUEST_*_CELL B\n"<<flush; 

	if(!cell->have()){
	  // The cell requested is actualy hold by another proc
	  // We need to get a cell back from the other proc
	  MPI_Bsend((void*)&cell->getIndex(),1,TYPE_CELLINDEX,cell->getActualId(),REQUEST_BACK_CELL,MPI_COMM_WORLD);
	  bool unique = true;
	  for(i=0;i<myRequestCell->count();i++)
	    if(myRequestCell->item(i) == index && myRequestId->item(i) == id){
	      unique = false;
	      //debug << "REQUEST_*_CELL done\n"<<flush; 
	      break;
	    }
	  if(unique){
	    myRequestId->prepend(id);
	    myRequestCell->prepend(index);
	  }
	  //debug << "REQUEST_*_CELL done\n"<<flush; 
	  break;
	}
	//debug << "REQUEST_*_CELL C\n"<<flush; 

	// We have the requested cell and send it
       	MPI_Bsend(cell->getItems(),(int)cell->count(),TYPE_PARTICLE,id,(tag == REQUEST_SRC_CELL) ?SEND_SRC_CELL:SEND_DEST_CELL,MPI_COMM_WORLD);
	cell->setActualId(id);
	if (cell->own())
	  // Cell goes away
	  myMissingCells->appendUnique(index);
	else
	  // Cell goes back home
	  myOtherCells->removeItem(index);

	//debug << "REQUEST_*_CELL done\n"<<flush; 
	break;

      case REQUEST_BACK_CELL:
	// There was a request for a cell which is need by an other proc than the owner
	// and we send it to the proc if possible
	//debug << "REQUEST_BACK_CELL\n"<<flush; 
	id = status.MPI_SOURCE;
	MPI_Recv(&index,1,TYPE_CELLINDEX,id,REQUEST_BACK_CELL,MPI_COMM_WORLD,&status);
	cell = getCell(index);
	already = (cell->getActualId() == id);

	if(already){
	  //debug << "REQUEST_BACK_CELL done\n"<<flush; 
	  break;
	}
	if(src == cell || dest == cell){
	  // We do not give away the cell which is needed for this interaction
	  bool unique = true;
	  for(i=0;i<myRequestCell->count();i++)
	    if(myRequestCell->item(i) == index && myRequestId->item(i) == id){
	      unique = false;
	      //debug << "REQUEST_BACK_CELL done\n"<<flush; 
	      break;
	    }
	  if(unique){
	    myRequestId->prepend(id);
	    myRequestCell->prepend(index);	  
	  }
	  //debug << "REQUEST_BACK_CELL done\n"<<flush; 
	  break;
	}

	MPI_Bsend(&index,1,TYPE_CELLINDEX,id,SEND_BACK_CELL,MPI_COMM_WORLD);
       	MPI_Bsend(cell->getItems(),(int)cell->count(),TYPE_PARTICLE,id,SEND_BACK_CELL,MPI_COMM_WORLD);

	cell->setActualId(id);
	// Cell goes back home
	myOtherCells->removeItem(index);
	//debug << "REQUEST_BACK_CELL done\n"<<flush; 

	break;
	
      case SEND_SRC_CELL:
	// We were asking for a cell and receive it
	//debug << "SEND_SRC_CELL\n"<<flush; 
	myCom++;
	index   = src->getIndex();
	id      = status.MPI_SOURCE;
	already = src->have();
	MPI_Probe(id,SEND_SRC_CELL,MPI_COMM_WORLD, &status);
	MPI_Get_count(&status,TYPE_PARTICLE,&count);

	if(already){
	  debug <<"already0\n";debug.flush();
	  MPI_Recv(NULL,0,TYPE_PARTICLE,id,SEND_SRC_CELL,MPI_COMM_WORLD, &status);
	}
	else
	  MPI_Recv(src->setItems(count),count,TYPE_PARTICLE,id,SEND_SRC_CELL,MPI_COMM_WORLD, &status);

	src->setActualId(myId);
	// Cell is back home
	myMissingCells->removeItem(index);

	//debug << "SEND_SRC_CELL done\n"<<flush; 
	break;
	
      case SEND_DEST_CELL:
	// We were asking for a cell and receive it
	
	//debug << "SEND_DEST_CELL\n"<<flush; 
	myCom++;
	index   = dest->getIndex();
	id      = status.MPI_SOURCE;
	already = dest->have();
	MPI_Probe(id,SEND_DEST_CELL,MPI_COMM_WORLD, &status);
	MPI_Get_count(&status,TYPE_PARTICLE,&count);

	if(already){
	  debug <<"already1\n";debug.flush();
	  MPI_Recv(NULL,0,TYPE_PARTICLE,id,SEND_DEST_CELL,MPI_COMM_WORLD, &status);
	}
	else 
	  MPI_Recv(dest->setItems(count),count,TYPE_PARTICLE,id,SEND_DEST_CELL,MPI_COMM_WORLD, &status);

	dest->setActualId(myId);
	if (dest->own())
	  // Cell is back home
	  myMissingCells->removeItem(index);
	else
	  // We got a cell from an other proc
	  myOtherCells->appendUnique(index);
	//debug << "SEND_DEST_CELL done\n"<<flush; 
	break;
	
      case SEND_BACK_CELL:
	// We were asking for a missing cell and receive it
	
	//debug << "SEND_BACK_CELL\n"<<flush; 
	myCom++;
	id      = status.MPI_SOURCE;
	MPI_Recv(&index,1,TYPE_CELLINDEX,id,SEND_BACK_CELL,MPI_COMM_WORLD,&status);
	cell    = getCell(index);
	already = cell->have();
	MPI_Probe(id,SEND_BACK_CELL,MPI_COMM_WORLD, &status);
	MPI_Get_count(&status,TYPE_PARTICLE,&count);

	if(already){
	  debug <<"already2\n";debug.flush();
	  MPI_Recv(NULL,0,TYPE_PARTICLE,id,SEND_BACK_CELL,MPI_COMM_WORLD, &status);
	}
	else
	  MPI_Recv(cell->setItems(count),count,TYPE_PARTICLE,id,SEND_BACK_CELL,MPI_COMM_WORLD, &status);

	cell->setActualId(myId);
	// Cell is back home
	myMissingCells->removeItem(index);

	if(src == cell || dest == cell){
	  //debug << "SEND_BACK_CELL done\n"<<flush; 
	  break;
	}

	for(i=myRequestId->count()-1;i>=0;i--){
	  if(index == myRequestCell->item(i)){
	    id = myRequestId->item(i);
	    already = (cell->getActualId() == id);

	    if(!cell->have())
	      continue;

	    myRequestCell->remove(i);
	    myRequestId->remove(i);

	    if(already)
	      continue;
	    
	    MPI_Bsend(cell->getItems(),(int)cell->count(),TYPE_PARTICLE,id,SEND_DEST_CELL,MPI_COMM_WORLD);
	    cell->setActualId(id);
	    myMissingCells->appendUnique(index);

	  }
	}
	//debug << "SEND_BACK_CELL done\n"<<flush; 
	break;
	
      case SEND_BACK_ALL_CELLS:
	// The proc has finished and sends my cells back
	//debug << "SEND_BACK_ALL_CELLS\n"<<flush; 
	id = status.MPI_SOURCE;
	MPI_Recv(&size,1,MPI_INT,id,SEND_BACK_ALL_CELLS,MPI_COMM_WORLD,&status);
	for(i=0;i<size;i++){
	  myCom++;
	  MPI_Recv(&index,1,TYPE_CELLINDEX,id,SEND_INDEX,MPI_COMM_WORLD,&status);
	  cell    = getCell(index);
	  already = (cell->have());
	  MPI_Probe(id,SEND_OTHER_CELL,MPI_COMM_WORLD, &status);
	  MPI_Get_count(&status,TYPE_PARTICLE,&count);
	  if(already){
	    debug <<"already3\n";debug.flush();
	    MPI_Recv(NULL,0,TYPE_PARTICLE,id,SEND_OTHER_CELL,MPI_COMM_WORLD, &status);
	  }
	  else
	    MPI_Recv(cell->setItems(count),count,TYPE_PARTICLE,id,SEND_OTHER_CELL,MPI_COMM_WORLD, &status);

	  cell->setActualId(myId);
	  myMissingCells->removeItem(index);

	  if (cell == src || cell == dest){
	    //debug << "Proc "<< myId <<": Receiving cell back "<<index<<", same as "<<((cell == src) ?"src":"dest")<<".\n";
	  }
	  else {
	    for(j=myRequestId->count()-1;j>=0;j--){
	      if(index == myRequestCell->item(j)){
		int id2 = myRequestId->item(j);
		already = (cell->getActualId() == id2);
		
		if(!cell->have())
		  continue;

		myRequestCell->remove(j);
		myRequestId->remove(j);
		
		if(already)
		  continue;
		
		MPI_Bsend(cell->getItems(),(int)cell->count(),TYPE_PARTICLE,id2,SEND_DEST_CELL,MPI_COMM_WORLD);
		cell->setActualId(id2);
		myMissingCells->appendUnique(index);
	      }
	    }
	  }
	}
	MPI_Bsend(NULL,0,MPI_CHAR,id,NO_MORE_CELLS,MPI_COMM_WORLD);
	myNeighboursCount--;
	//debug << "SEND_BACK_ALL_CELLS done\n"<<flush; 
	break;
	
      case NO_MORE_CELLS:
	//debug << "NO_MORE_CELLS\n"<<flush; 
	id = status.MPI_SOURCE;
	MPI_Recv(NULL,0,MPI_CHAR,id,NO_MORE_CELLS,MPI_COMM_WORLD,&status);
	backReq--;
	//debug << "NO_MORE_CELLS done\n"<<flush; 
	break;
      default:
	cerr <<"Gulp! I got a wrong msg with tag "<<status.MPI_TAG<<" from "<<status.MPI_SOURCE<<".\n";
	break;
	
      }

    }
    if(reqSrc && src->have()){
      reqSrc = false;
      flag = reqDest || reqSrc;
    }
    if(reqDest && dest->have()){
      reqDest = false;
      flag = reqDest || reqSrc;
    }
    if(src != NULL && dest != NULL && (!src->have() || !dest->have())){
      if(!flag)
	cerr <<"Proc "<<myId<<" had the cells ...\n";
      flag = true;
    }
  }
  //    if(myRequestId->count()>0 && src == NULL && dest == NULL){
  //      cerr << "Proc "<<myId<< " has still "<<myRequestId->count()<<" request!\n";
  //      for(i=0;i<myRequestId->count();i++)
  //        cerr <<(CellIndex&)myRequestCell->item(i)<<","<<myRequestId->item(i)<<","<<getCell(myRequestCell->item(i))->getActualId()<<"\n"; 
  //    }
  //    if(myOtherCells->count()>0 && src == NULL && dest == NULL){
  //      cerr << "Proc "<<myId<< " has still "<<myOtherCells->count()<<" others!\n";
  //      for(i=0;i<myOtherCells->count();i++)
  //        cerr <<(CellIndex&)myOtherCells->item(i)<<"\n"; 
  //    }
  //    if(myMissingCells->count()>0 && src == NULL && dest == NULL){
  //      cerr << "Proc "<<myId<< " has still "<<myMissingCells->count()<<" missings!\n";
  //      for(i=0;i<myMissingCells->count();i++)
  //        cerr <<(CellIndex&)myMissingCells->item(i)<<"\n"; 
  //    }
  //    if(src != NULL && dest != NULL && (!src->have() || !dest->have()))
  //      cerr <<"Proc "<<myId<<" src "<<src->getActualId()<<" and dest "<<dest->getActualId()<<"!\n";

  myRequestList->removeAll();
  //debug <<"sendRecvCell done\n"<<flush;

  myTimeCom += mytime()-com-overheadtime;
}
#else
void Universe::sendRecvCell(Cell*, Cell*){}
#endif

#ifdef HAVE_MPI
void Universe::sendRecvCellSimple(long i){
  if (numProc < 2) return;
  
  long j;
  Cell* cell = NULL;
  
  double com = mytime();
  if(i<0){
    // We have to send back all not owned cells and receive our calls
    MPI_Waitall((int)myRequestList->count(),myRequestList->getItems(),myStatusList->getItems());
    myRequestList->removeAll();
    myStatusList->removeAll();

    for(j=0;j<myRecvList->count();j++){
      cell = getCell(myRecvList->item(j));
      MPI_Request request;
      MPI_Status status;
      Cell* cell = getCell(myRecvList->item(j));
      MPI_Isend(cell->getItems(),(int)cell->count(),TYPE_PARTICLE,cell->getOwnerId(),LLA_CELLS,MPI_COMM_WORLD,&request);
      myRequestList->add(request);
      myStatusList->add(status);
      //cell->setActualId(cell->getOwnerId());
      myCom++;
    }

    for(j=0;j<mySendList->count();j++){
      MPI_Status status;
      cell = getCell(mySendList->item(j));
      MPI_Recv(cell->setItems(cell->count()),(int)cell->count(),TYPE_PARTICLE,cell->usedBy(0),LLA_CELLS,MPI_COMM_WORLD, &status);
      //cell->setActualId(myId);
    }

    MPI_Waitall((int)myRequestList->count(),myRequestList->getItems(),myStatusList->getItems());
    myRequestList->removeAll();
    myStatusList->removeAll();
    MPI_Barrier(MPI_COMM_WORLD);
  }
  else if(i == myN+1){
    // We send the cells
    //cerr << "Proc "<<myId<< " send at "<<i<<".\n";
    myRequestList->removeAll();
    myStatusList->removeAll();
    for(j=0;j<mySendList->count();j++){
      MPI_Request request;
      MPI_Status status;
      cell = getCell(mySendList->item(j));
      MPI_Isend(cell->getItems(),(int)cell->count(),TYPE_PARTICLE,cell->usedBy(0),ALL_CELLS,MPI_COMM_WORLD,&request);

      myRequestList->add(request);
      myStatusList->add(status);
      //cell->setActualId(cell->usedBy(0));
      myCom++;
    }
  }
  else if(i ==  myOwnCells->count() - myN -3){
    // We reciev the cells we need
    //cerr << "Proc "<<myId<<" recv at "<<i<<".\n";

    for(j=0;j<myRecvList->count();j++){
      cell = getCell(myRecvList->item(j));
      int count;
      MPI_Status status;
      MPI_Probe(cell->getOwnerId(),ALL_CELLS,MPI_COMM_WORLD, &status);
      MPI_Get_count(&status,TYPE_PARTICLE,&count);      
      MPI_Recv(cell->setItems(count),count,TYPE_PARTICLE,cell->getOwnerId(),ALL_CELLS,MPI_COMM_WORLD, &status);
      //cell->setActualId(myId);
    }
  }

  myTimeCom += mytime()-com-overheadtime;
}
#else
void Universe::sendRecvCellSimple(long){}
#endif

#ifdef HAVE_MPI
void Universe::sendRecvCellSeq(bool flag){
  if (numProc < 2) return;
  
  long j;
  Cell* cell = NULL;
  
  double com = mytime();

  if(flag){
    if(myId == master){
      //cerr << myId <<": Recv-recv\n";
      for(j=0;j<myRecvList->count();j++){
	cell = getCell(myRecvList->item(j));
	int count;
	MPI_Status status;
	MPI_Probe(cell->getOwnerId(),ALL_CELLS,MPI_COMM_WORLD, &status);
	MPI_Get_count(&status,TYPE_PARTICLE,&count);      
	MPI_Recv(cell->setItems(count),count,TYPE_PARTICLE,cell->getOwnerId(),ALL_CELLS,MPI_COMM_WORLD, &status);
      }
    }
    else {
      //cerr << myId <<": Send-send\n";
      for(j=0;j<mySendList->count();j++){
	cell = getCell(mySendList->item(j));
	MPI_Send(cell->getItems(),(int)cell->count(),TYPE_PARTICLE,cell->usedBy(0),ALL_CELLS,MPI_COMM_WORLD);
	myCom++;
      }
      //cerr << myId <<": Recv-send\n";
      for(j=0;j<mySendList->count();j++){
	MPI_Status status;
	cell = getCell(mySendList->item(j));
	MPI_Recv(cell->setItems(cell->count()),(int)cell->count(),TYPE_PARTICLE,cell->usedBy(0),LLA_CELLS,MPI_COMM_WORLD, &status);
      }
      //cerr << myId <<": Recv-recv\n";
      for(j=0;j<myRecvList->count();j++){
	cell = getCell(myRecvList->item(j));
	int count;
	MPI_Status status;
	MPI_Probe(cell->getOwnerId(),ALL_CELLS,MPI_COMM_WORLD, &status);
	MPI_Get_count(&status,TYPE_PARTICLE,&count);      
	MPI_Recv(cell->setItems(count),count,TYPE_PARTICLE,cell->getOwnerId(),ALL_CELLS,MPI_COMM_WORLD, &status);
      }
    }
  }
  else {
    if(myId == master){
      //cerr << myId <<": Send-recv\n";
      for(j=0;j<myRecvList->count();j++){
	cell = getCell(myRecvList->item(j));
	Cell* cell = getCell(myRecvList->item(j));
	MPI_Send(cell->getItems(),(int)cell->count(),TYPE_PARTICLE,cell->getOwnerId(),LLA_CELLS,MPI_COMM_WORLD);
	myCom++;
      }
      //cerr << myId <<": Send-send\n";
      for(j=0;j<mySendList->count();j++){
	cell = getCell(mySendList->item(j));
	MPI_Send(cell->getItems(),(int)cell->count(),TYPE_PARTICLE,cell->usedBy(0),ALL_CELLS,MPI_COMM_WORLD);
	myCom++;
      }
      //cerr << myId <<": Recv-send\n";
      for(j=0;j<mySendList->count();j++){
	MPI_Status status;
	cell = getCell(mySendList->item(j));
	MPI_Recv(cell->setItems(cell->count()),(int)cell->count(),TYPE_PARTICLE,cell->usedBy(0),LLA_CELLS,MPI_COMM_WORLD, &status);
      }
    }
    else{
      //cerr << myId <<": Send-recv\n";
      for(j=0;j<myRecvList->count();j++){
	cell = getCell(myRecvList->item(j));
	Cell* cell = getCell(myRecvList->item(j));
	MPI_Send(cell->getItems(),(int)cell->count(),TYPE_PARTICLE,cell->getOwnerId(),LLA_CELLS,MPI_COMM_WORLD);
	myCom++;
      }
    }
  }
  myTimeCom += mytime()-com-overheadtime;
  
}
#else
void Universe::sendRecvCellSeq(bool){}
#endif




void Universe::dump(){

  long j;

  if((myDtCount % stat_number) == 0 && myDtCount> 0){
    Real d3[] = {0.0, 0.0, 0.0};
    for(j=0;j<stat_number;j++){
      d3[0] += myStat[j][0]/stat_number; 
      d3[1] += myStat[j][1]/stat_number; 
      d3[2] += myStat[j][2]/stat_number; 
    }
    if(myId == master){
      //cerr << myTime <<"\t"<<d3[0]<<"\t"<<d3[1]<<"\t"<<d3[0]+d3[1]<<"\t"<<d3[2]<<"\t"<<(myL1-myL0)/myL0<<"\t"<<myL1<<"\n"; 
      output << myTime <<" "<<d3[0]<<" "<<d3[1]<<" "<<d3[0]+d3[1]<<" "<<d3[2]<<" "<<(myL1-myL0)/myL0<<" "<<myL1<<"\n";
      output.flush();
    }
  }
  //if(myId == master)
  //  cerr << myTime <<"\t"<<myEkin<<"\t"<<myEpot<<"\t"<<myFa <<"\n";

  myStat[myDtCount % stat_number][0] = myEkin;
  myStat[myDtCount % stat_number][1] = myEpot;
  myStat[myDtCount % stat_number][2] = myFa;

  if(!myDump || myDumpTab[myDumpCount] > myTime)
    return;


  long count = myOwnCells->count();
  long i;
  char filename[buffer_len];



  if(numProc > 1){
    sprintf(filename,"%s.%04d.%03d",myOutFilename,(int)myDumpCount,myId);
  }
  else{
    sprintf(filename,"%s.%04d",myOutFilename,(int)myDumpCount);
  }
  ofstream dumpFile(filename, ios::out);

  //cerr << "Proc "<<myId << ", dump at "<<myTime<<", "<<count<<" cells\n";
  if(myId == master)
    dumpFile <<myTime<<"\n";
  for(i=0;i<count;i++){
    for(j=0;j<getCell(myOwnCells->item(i))->count();j++){
      const Particle& par = getCell(myOwnCells->item(i))->item(j);
      dumpFile << par.tag+1 << " "
	       << par.x    << " "
	       << par.y    << " "
	       << par.vx   << " "
	       << par.vy   << " "   
	       << par.epot << "\n";   
    }
  }
  dumpFile.close();
  myDumpCount++;
}

long Universe::getI(Real x) const {
  long i = (long)((x-myLeft)*myLxN);
  /*if(periodic_boundary_x)
    i = (i+myN) % myN;
    else*/ if (i < 0)
      i = 0;
    else if (i >= myN)
      i = myN-1;
  return i;
}

long Universe::getJ(Real y) const {
  long j = (long)((y-myBottom)*myLyM);
  /*if(periodic_boundary_y)
    j = (j+myM) % myM;
    else*/ if (j < 0)
      j = 0;
    else if (j >= myM)
      j = myM-1;
  return j;
}

FastList<Particle>* Universe::getParticles(){

  myTmpList->removeAll();
  long count = myOwnCells->count();

  for(long i=0;i<count;i++)
    myTmpList->addFast(getCell(myOwnCells->item(i))->count(),getCell(myOwnCells->item(i))->getItems());

  return myTmpList;
}

FastList<Particle>* Universe::getParticlesFromAll(){

#ifdef HAVE_MPI
  if(numProc<2)
    return getParticles();
  bool ask = true;
  MPI_Status status;
  myRequestArray[master] = MPI_REQUEST_NULL;
  for(int i=0;i<numProc;i++)
    if(myRequestArray[i] != MPI_REQUEST_NULL){
      ask = false;
      break;
    }

  if(ask){
    for(int i=0;i<numProc;i++)
      if(master != i)
	MPI_Isend(NULL,0,TYPE_PARTICLE,i,REQUEST_PARTICLES,MPI_COMM_WORLD,&myRequestArray[i]);
    return getParticles();
  }

  int count;
  MPI_Probe(MPI_ANY_SOURCE,SEND_PARTICLES,MPI_COMM_WORLD, &status);
  MPI_Get_count(&status,TYPE_PARTICLE,&count);
  int id = status.MPI_SOURCE;
  MPI_Recv(myTmpList->setItems(count),count,TYPE_PARTICLE,id,SEND_PARTICLES,MPI_COMM_WORLD, &status);
  MPI_Wait(&myRequestArray[id],&status);
  return myTmpList;
#else
  return getParticles();
#endif
}

void Universe::serveDisplay(){
#ifdef HAVE_MPI
  if(myId == master){
    myRequestArray[master] = MPI_REQUEST_NULL;
    for(int i=0;i<numProc;i++)
      if(i != master)
	MPI_Isend(NULL,0,MPI_CHAR,i,NO_MORE_PARTICLES,MPI_COMM_WORLD,&myRequestArray[i]);
    MPI_Waitall(numProc,myRequestArray,myStatusArray);
  }
  else{
    bool quit = false;
    while(!quit){
      MPI_Status status;
      FastList<Particle>* list = NULL;

      MPI_Probe(master,MPI_ANY_TAG,MPI_COMM_WORLD, &status);
      switch (status.MPI_TAG) {
      case NO_MORE_PARTICLES:
      	MPI_Recv(NULL,0,MPI_CHAR,master,NO_MORE_PARTICLES,MPI_COMM_WORLD,&status);
	quit = true;
	break;
      case REQUEST_PARTICLES:
	list = getParticles();
      	MPI_Recv(NULL,0,MPI_CHAR,master,REQUEST_PARTICLES,MPI_COMM_WORLD,&status);
	MPI_Send(list->getItems(),(int)list->count(),TYPE_PARTICLE,master,SEND_PARTICLES,MPI_COMM_WORLD);
	break;
      default:
	cerr <<"[Universe::serveDisplay] Gulp! I got a wrong msg:"<<status.MPI_TAG<<", "<<status.MPI_SOURCE<<"\n";
	break;
      }
    }
  }
#endif
}



void Universe::pushSharedCells(){
#ifdef HAVE_ONE
  if(numProc<2)
    return;
  
  double com = mytime();
  long i,j;
  MPI_Aint start;
  MPI_Address(myShmem[myId],&start);
#ifndef BLOCKING_GET
  MPI_Win_lock(MPI_LOCK_EXCLUSIVE,myId,0,myWindow[myId]);
#else
  for(j=0;j<numProc;j++)
    MPI_Win_fence(0,myWindow[j]);
#endif
  for(i=0;i<myNumberSharedCells[myId];i++){
    Cell* cell = getCell(mySharedCells->item(i)); 
    MPI_Aint disp;
    MPI_Address(cell->getShmPtr(),&disp);
    disp = disp - start;
    myCom++;
    MPI_Put(cell->getPtr(),1,TYPE_CELL,myId,disp,1,TYPE_CELL,myWindow[myId]);
  }
#ifndef BLOCKING_GET
  MPI_Win_unlock(myId,myWindow[myId]);
  // Make sure that we do not start to access before put is executed
  MPI_Barrier(MPI_COMM_WORLD);
#else
  for(j=0;j<numProc;j++)
    MPI_Win_fence(0,myWindow[j]);
#endif

  double t = mytime()-com-overheadtime;
  myTimeCom += t;
  myTimeComPush += t;
  
#endif
}

void Universe::pushSharedCellsFast(){
#ifdef HAVE_ONE
  if(numProc<2)
    return;

  
  double com = mytime();
  long i;
  MPI_Aint start,disp;
  MPI_Address(myShmem[myId],&start);

  //MPI_Barrier(MPI_COMM_WORLD);
  //for(i=0;i<numProc;i++)
  //  MPI_Win_fence(0,myWindow[i]);

#ifndef BLOCKING_GET
  MPI_Win_lock(MPI_LOCK_SHARED,myId,0,myWindow[myId]);
#else
  for(i=0;i<numProc;i++)
    MPI_Win_fence(0,myWindow[i]);
#endif

  for(i=0;i<myNumberSharedCells[myId];i++){
    Cell* cell = getCell(mySharedCells->item(i)); 
    MPI_Aint disp;
    MPI_Address(cell->getShmPtr(),&disp);
    disp = disp - start;
    myCom++;
    MPI_Put(cell->getPtr(),1,TYPE_CELL,myId,disp,1,TYPE_CELL,myWindow[myId]);
  }

#ifndef BLOCKING_GET
  MPI_Win_unlock(myId,myWindow[myId]);
#endif  

  for(i=0;i<numProc;i++)
    MPI_Win_fence(0,myWindow[i]);

  long count = myOtherCellsUsed->count(); 
  for(i=0;i<count;i++){
    Cell* cell = getCell(myOtherCellsUsed->item(i));
    MPI_Address(myShmem[cell->getOwnerId()],&start);
    MPI_Address(cell->getShmPtr(),&disp);
    disp = disp - start;
    //MPI_Win_lock(MPI_LOCK_SHARED,cell->getOwnerId(),0,myWindow[cell->getOwnerId()]);
    MPI_Get(cell->getPtr(),1,TYPE_CELL,cell->getOwnerId(),disp,1,TYPE_CELL,myWindow[cell->getOwnerId()]);
    //MPI_Win_unlock(cell->getOwnerId(),myWindow[cell->getOwnerId()]);
    myCom++;
  }

  for(i=0;i<numProc;i++)
    MPI_Win_fence(0,myWindow[i]);

  double t = mytime()-com-overheadtime;
  myTimeCom += t;
  myTimeComPush += t;
  
#endif
}

#ifdef HAVE_ONE
void Universe::getRemoteCell(Cell* cell){
  if(numProc<2 || cell->have() || cell->own())
    return;
  
  double com = mytime();

#ifndef BLOCKING_GET
  MPI_Win_lock(MPI_LOCK_SHARED,cell->getOwnerId(),0,myWindow[cell->getOwnerId()]);
#endif
  MPI_Aint start,disp;
  MPI_Address(myShmem[cell->getOwnerId()],&start);
  MPI_Address(cell->getShmPtr(),&disp);
  disp = disp - start;

  MPI_Get(cell->getPtr(),1,TYPE_CELL,cell->getOwnerId(),disp,1,TYPE_CELL,myWindow[cell->getOwnerId()]);

  cell->setActualId(myId);
  myOtherCells->appendUnique(cell->getIndex());

  myCom++;

#ifndef BLOCKING_GET
  MPI_Win_unlock(cell->getOwnerId(),myWindow[cell->getOwnerId()]);
#endif

  double t = mytime()-com-overheadtime;
  myTimeCom += t;
  myTimeComGet += t;
  


}
#else
void Universe::getRemoteCell(Cell*){}
#endif


void Universe::popSharedCells(){
#ifdef HAVE_ONE
  if(numProc<2)
    return;
  
  double com = mytime();
  long i,p;
  // Wait for all proc
  // We need this barrier to avoid that some nodes stil
  // request data from others.
  MPI_Barrier(MPI_COMM_WORLD);
  for(i=0;i<numProc;i++)
    MPI_Win_fence(0,myWindow[i]);
  //myTimeIdle += mytime()-idle;

  // Write the forces to shared
  int count = (int)myOtherCells->count(); 
  for(i=0;i<count;i++){
    Cell* cell = getCell(myOtherCells->item(i)); 
    MPI_Aint start,disp;
    MPI_Address(myShmem[cell->getOwnerId()],&start);
    MPI_Address(cell->getShmPtr(),&disp);
    disp = disp - start;
    //MPI_Win_lock(MPI_LOCK_SHARED,myId,0,myWindow[cell->getOwnerId()]);
    MPI_Put(cell->getPtr(),1,TYPE_CELL,myId,disp,1,TYPE_CELL,myWindow[cell->getOwnerId()]);
    //MPI_Win_unlock(myId,myWindow[cell->getOwnerId()]);
    cell->setActualId(cell->getOwnerId());
    myCom++;
  }
  myOtherCells->removeAll();

  // Wait for all proc
  for(i=0;i<numProc;i++)
    MPI_Win_fence(0,myWindow[i]);


  // Accumulate the forces from other procs
#if 1
  // Fastest, moving the whole window of cells
  count = (int)mySharedCells->count(); 
  for(p=0;p<myNeighbours->count();p++){
    int id = myNeighbours->item(p);
    myCom += count;
#ifndef BLOCKING_GET
    MPI_Win_lock(MPI_LOCK_SHARED,id,0,myWindow[myId]);
#endif
    MPI_Get(myTmpLocalCells,count,TYPE_CELL,id,0,count,TYPE_CELL,myWindow[myId]);
#ifndef BLOCKING_GET
    MPI_Win_unlock(id,myWindow[myId]);
#endif
    char* ptr = (char*)myTmpLocalCells+mymallocsize(Particle,0);
    for(i=0;i<count;i++){
      Cell*     cell = getCell(mySharedCells->item(i));
      if(cell->usedById(id)){
	long count2 = cell->count();
	Particle* par  = cell->getItems();
	Particle* tmp  = (Particle*)ptr;
	for(long j=0;j<count2;j++){
	  par[j].fxa  += tmp[j].fxa;
	  par[j].fya  += tmp[j].fya;
	  par[j].epot += tmp[j].epot;
	}
      }
      ptr +=mymallocsize(Particle,myMaxPerCell);
    }
  }

#ifdef BLOCKING_GET
  for(i=0;i<numProc;i++)
    MPI_Win_fence(0,myWindow[i]);
#endif

#endif
#if 0
  // Faster, moving a complete cell
  count = (int)mySharedCells->count(); 
  MPI_Aint start;
  MPI_Address(myShmem[myId],&start);

  for(i=0;i<count;i++){
    Cell*     cell = getCell(mySharedCells->item(i));
    long count2 = cell->count();
    Particle* par  = cell->getItems();
    myCom++;

    myTmpList->removeAll();
    myTmpList->atLeastSize(myMaxPerCell);

    MPI_Aint disp;
    MPI_Address(cell->getShmPtr(),&disp);
    disp = disp - start;
    for(long k=0;k<max_neighbours;k++){
      if(cell->usedBy(k) < 0)
	break;
#ifndef BLOCKING_GET
      MPI_Win_lock(MPI_LOCK_SHARED,cell->usedBy(k),0,myWindow[myId]);
#endif
      MPI_Get(myTmpList->getPtr(),1,TYPE_CELL,cell->usedBy(k),disp,1,TYPE_CELL,myWindow[myId]);
#ifndef BLOCKING_GET
      MPI_Win_unlock(cell->usedBy(k),myWindow[myId]);
#endif
      Particle* tmp  = myTmpList->getItems();
      for(long j=0;j<count2;j++){
	par[j].fxa  += tmp[j].fxa;
	par[j].fya  += tmp[j].fya;
	par[j].epot += tmp[j].epot;
      }
    }
  }
#ifdef BLOCKING_GET
  for(i=0;i<numProc;i++)
    MPI_Win_fence(0,myWindow[i]);
#endif

#endif
#if 0
  // Orignial ... slow version, for some!
  // Slow, moving just one particle
  long offsetA;
  long offsetB;
  {
    Particle par[2];
    MPI_Aint start,tmp,end;
    MPI_Address(&par[0],&start);
    MPI_Address(&(par[0].fxa),&tmp);
    MPI_Address(&par[1],&end);
    
    offsetA = end-start;
    offsetB = 2*(long)sizeof(long) + (tmp - start);
  }
  count = (int)mySharedCells->count(); 
  for(i=0;i<count;i++){
    Cell*     cell = getCell(mySharedCells->item(i));
    Particle* par  = cell->getItems();
    myCom++;

    MPI_Aint start,disp;
    MPI_Address(myShmem[myId],&start);
    MPI_Address(cell->getShmPtr(),&disp);
    disp = disp - start + offsetB;
    long count2 = cell->count();
    for(long j=0;j<count2;j++){
      for(long k=0;k<max_neighbours;k++){
	if(cell->usedBy(k) < 0)
	  break;
	Real tmp[3] = {0,0,0};
#ifndef BLOCKING_GET
	MPI_Win_lock(MPI_LOCK_SHARED,cell->usedBy(k),0,myWindow[myId]);
#endif
	MPI_Get(tmp,3,TYPE_REAL,cell->usedBy(k),disp,3,TYPE_REAL,myWindow[myId]);
#ifndef BLOCKING_GET
	MPI_Win_unlock(cell->usedBy(k),myWindow[myId]);
#endif
	par[j].fxa  += tmp[0];
	par[j].fya  += tmp[1];
	par[j].epot += tmp[2];
      }
      disp += offsetA;
    }
  }

#ifdef BLOCKING_GET
  for(i=0;i<numProc;i++)
    MPI_Win_fence(0,myWindow[i]);
#endif

#endif

  double t = mytime()-com-overheadtime;
  myTimeCom += t;
  myTimeComPop += t;
  
#endif
}

void Universe::popSharedCellsFast(){
#ifdef HAVE_ONE
  if(numProc<2)
    return;
  
  double com = mytime();
  long i,p;
  // Wait for all proc
  MPI_Barrier(MPI_COMM_WORLD);
  for(i=0;i<numProc;i++)
    MPI_Win_fence(0,myWindow[i]);

  // Write the forces to shared
  int count = (int)myOtherCellsUsed->count(); 
  for(i=0;i<count;i++){
    Cell* cell = getCell(myOtherCellsUsed->item(i)); 
    MPI_Aint start,disp;
    MPI_Address(myShmem[cell->getOwnerId()],&start);
    MPI_Address(cell->getShmPtr(),&disp);
    disp = disp - start;
    //MPI_Win_lock(MPI_LOCK_SHARED,myId,0,myWindow[cell->getOwnerId()]);
    MPI_Put(cell->getPtr(),1,TYPE_CELL,myId,disp,1,TYPE_CELL,myWindow[cell->getOwnerId()]);
    //MPI_Win_unlock(myId,myWindow[cell->getOwnerId()]);
    cell->setActualId(cell->getOwnerId());
    myCom++;
  }
  for(i=0;i<numProc;i++)
    MPI_Win_fence(0,myWindow[i]);

  // Wait for all proc
#ifndef BLOCKING_GET
  MPI_Barrier(MPI_COMM_WORLD);
#endif

  // Accumulate the forces from other procs
  count = (int)mySharedCells->count(); 
  for(p=0;p<myNeighbours->count();p++){
    int id = myNeighbours->item(p);
    myCom += count;
#ifndef BLOCKING_GET
    MPI_Win_lock(MPI_LOCK_SHARED,id,0,myWindow[myId]);
#endif
    MPI_Get(myTmpLocalCells,count,TYPE_CELL,id,0,count,TYPE_CELL,myWindow[myId]);
#ifndef BLOCKING_GET
    MPI_Win_unlock(id,myWindow[myId]);
#endif
    char* ptr = (char*)myTmpLocalCells+mymallocsize(Particle,0);
    for(i=0;i<count;i++){
      Cell*     cell = getCell(mySharedCells->item(i));
      if(cell->usedById(id)){
	long count2 = cell->count();
	Particle* par  = cell->getItems();
	Particle* tmp  = (Particle*)ptr;
	for(long j=0;j<count2;j++){
	  par[j].fxa  += tmp[j].fxa;
	  par[j].fya  += tmp[j].fya;
	  par[j].epot += tmp[j].epot;
	}
      }
      ptr +=mymallocsize(Particle,myMaxPerCell);
    }
  }

#ifdef BLOCKING_GET
  for(i=0;i<numProc;i++)
    MPI_Win_fence(0,myWindow[i]);
#endif

  double t = mytime()-com-overheadtime;
  myTimeCom += t;
  myTimeComPop += t;
  
#endif
}

#ifdef HAVE_ONE
void  Universe::build_derived_type_Cell(MPI_Datatype *handle, long size){

  //if(myId == master)
  //  cerr <<"[Universe::build_derived_type_Cell]\n";

  int block_lengths[] = { (int)mymallocsize(Particle,size), 1};
  MPI_Datatype types[] = { MPI_BYTE, MPI_UB }; 
  MPI_Aint displacements[2];
  
  displacements[0] = 0;
  displacements[1] = mymallocsize(Particle,size);
  
  MPI_Type_struct(2, block_lengths, displacements, types, handle);
  MPI_Type_commit(handle);  

  myTypeCount += 2;

}

#endif
