/*
 * Universe.h
 *
 * Copyright (C) 1999-2000, Thierry Matthey <matthey@ii.uib.no>
 *
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#ifndef _UNIVERSE_H
#define _UNIVERSE_H

#ifdef NO_ANSI_CPP
#include <iostream.h>
#include <fstream.h>
#else
#include <iostream>
#include <fstream>
using std::cerr;
using std::ofstream;
using std::ostream;
using std::ios;
#endif

#include "global.h"
#include "Cell.h"

class StartWorld;
class Force;

//_________________________________________________________________ Universe

class Universe {

 public:
enum CommunicationType {
  CommunicationDistributed  = 0,
  CommunicationOneSided     = 1,
  CommunicationSimple       = 2,
  CommunicationSequential   = 3,
  CommunicationOneSidedFast = 4
};

enum PartitionType {
  PartitionStripe    = 0,
  PartitionMetis     = 1,
  PartitionChecker   = 2,
  PartitionRectangle = 3
};
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Constructors, destructors, assignment
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 public:
  
  Universe(StartWorld* world, CommunicationType com, PartitionType part, bool dumpflag);
  ~Universe();
 private:
  Universe& operator=(const Universe&);

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// New methods of class Universe
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 public:
  friend ostream& operator<<(ostream& OS, Universe & universe);

  void next();
  void serveDisplay();
  void display();
  bool isDone() const;
  void dump();

  Cell* getCell(long i, long j) const;
  Cell* getCell(Real x, Real y) const;
  Cell* getCell(const CellIndex& index) const;
  Cell* getCell(const Particle& par) const;

  int getOwnerId(Real x, Real y) const;
  int getOwnerId(long i, long j) const;
  int getOwnerId(const Particle& par) const;
  int getOwnerId(const CellIndex& index) const;

  int getActualId(Real x, Real y) const;
  int getActualId(long i, long j) const;
  int getActualId(const Particle& par) const;
  int getActualId(const CellIndex& index) const;

  CellIndex getIndex(Real x, Real y) const;
  CellIndex getIndex(const Particle& par) const;
  CellIndex getIndex(long i) const;

  void addParticle(const Particle& particle);

  FastList<Particle>* getParticles();
  FastList<Particle>* getParticlesFromAll();

  Real getLeft() const;
  Real getRight() const;
  Real getLx() const;
  Real getTop() const;
  Real getBottom() const;
  Real getLy() const;
  long getNStop() const;
  Real getTStop() const;
  long getCount() const;
  Real getTime() const;
  Real getDt() const;
  void setL0(Real l0);
  long getNumberTotal() const;


 private:
  void partitioning();
  void neighbourhood();
  void computeInteraction(Cell* src, Cell* dest, Real bx = 0.0, Real by = 0.0);
  void force();
  void update();
  void sendRecvCell(Cell* src, Cell* dest);
  void sendRecvCellSimple(long i);
  void sendRecvCellSeq(bool flag);

  long getI(Real x) const;
  long getJ(Real y) const;

#ifdef HAVE_ONE
  void build_derived_type_Cell(MPI_Datatype *handle, long size);
#endif  
  void pushSharedCells();
  void getRemoteCell(Cell* cell);
  void popSharedCells();
  void pushSharedCellsFast();
  void popSharedCellsFast();

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// My data members
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 private:  

  Force*           myForce;          // force
  Cell*            myCells;          // array with all cells
  List<CellIndex>* myOwnCells;       // list of the owned cells
  List<CellIndex>* myOtherCells;     // temp list of cells from other procs
  List<CellIndex>* myOtherCellsUsed; // list of cells used owend by other procs
  List<CellIndex>* myMissingCells;   // temp list of cells borrowed to other procs

  long myN;    
  Real myLeft;
  Real myRight;
  Real myLx;
  Real myLxN;
  Real myA;

  long myM;
  Real myTop;
  Real myBottom;
  Real myLy;
  Real myLyM;
  Real myB;

  bool myDump;
  Universe::CommunicationType myCommunication;
  Universe::PartitionType     myPartition;
  long myCount;
  long myDtCount;
  long myNStop;
  long myDumpCount;
  long myNumber;
  long myNumberTotal;
  long myInteractions;
  long myMaxPerCell;
  Real myTStop;
  Real myTime;
  Real myDt;
  Real myDj;
  Real myRCut;
  Real myTemp;
  Real myPullStart;
  Real myPullStop;
  Real myFa;

  Real myDtSqrt;
  Real myDt2;
  Real myDthalf;
  Real myKsiT;
  Real myAlfa;
  Real myKsi;
  Real myTStep;
  Real myEkin;
  Real myEpot;
  Real myL0;
  Real myL1;
  Real myDensity;
  bool mySizeTest;

  double myTimeTotal;
  double myTimeSim;
  double myTimeCalc;
  double myTimeCom;
  double myTimeComPush;
  double myTimeComPop;
  double myTimeComGet;

  double myTimeIdle;
  double myTimeUpdate;
  double myTimeDist;
  double myTimeRest;

  long myCom;
  long myComInter;
  long myComBrute;
  long myDistribute;
  long myWeightSum;
  long long myParInter;

  char myOutFilename[buffer_len];
  char myInFilename[buffer_len];
  Real myStat[max_stat_values][3];

  List<int>* myNeighbours;
  List<int>* myOverlap;
  long       myNeighboursCount;

  List<CellIndex>* myRequestCell;
  List<int>*       myRequestId;

  FastList<Particle>* myTmpList;
  FastList<Particle>* myUpdateParticles;
  List<CellIndex>*    mySendList;
  List<CellIndex>*    myRecvList;

  int* myIntArray;
  int* myIntArray2;
  Real* myDumpTab;
  Real* myBoundsX;
  Real* myBoundsY;


#ifdef HAVE_MPI
  FastList<MPI_Request>* myRequestList;
  FastList<MPI_Status>* myStatusList;

  MPI_Request* myRequestArray;
  MPI_Status* myStatusArray;

  void*            myBuffer;
  int              myNBuffer;
#endif

#ifdef HAVE_ONE 
  List<CellIndex>* mySharedCells;

  void**           myShmem;
  MPI_Win*         myWindow;
  long*            myNumberSharedCells;

  MPI_Datatype     TYPE_CELL;
  long             myTypeCount;
  void*            myTmpLocalCells;
#endif

};

//______________________________________________________________________ INLINES

inline Cell* Universe::getCell(const CellIndex& index) const{
  return &myCells[index.i()+index.j()*myN];
}
inline Cell* Universe::getCell(long i, long j) const{
  return &myCells[i+j*myN];
}
inline Cell* Universe::getCell(const Particle& par) const{
  return &myCells[getI(par.x)+getJ(par.y)*myN];
}
inline Cell* Universe::getCell(Real x, Real y) const{
  return &myCells[getI(x)+getJ(y)*myN];
}
inline CellIndex Universe::getIndex(const Particle& par) const{
  return CellIndex(getI(par.x),getJ(par.y));
}
inline CellIndex Universe::getIndex(Real x, Real y) const{
  return CellIndex(getI(x),getJ(y));
}
inline CellIndex Universe::getIndex(long i) const{
  return CellIndex(i%myN,i/myN);
}
inline int Universe::getOwnerId(Real x, Real y) const{
  return (getCell(x,y)->getOwnerId());
}
inline int Universe::getOwnerId(long i, long j) const{
  return (getCell(i,j)->getOwnerId());
}
inline int Universe::getOwnerId(const CellIndex& index) const{
  return (getCell(index)->getOwnerId());
}
inline int Universe::getOwnerId(const Particle& par) const{
  return (getCell(par.y,par.x)->getOwnerId());
}
inline int Universe::getActualId(Real x, Real y) const{
  return (getCell(x,y)->getActualId());
}
inline int Universe::getActualId(long i, long j) const{
  return (getCell(i,j)->getActualId());
}
inline int Universe::getActualId(const CellIndex& index) const{
  return (getCell(index)->getActualId());
}
inline int Universe::getActualId(const Particle& par) const{
  return (getCell(par.y,par.x)->getActualId());
}

inline void Universe::addParticle(const Particle& particle){
  getCell(particle)->add(particle);
}
inline bool Universe::isDone() const{
  return (myTime + myDt/2 > myTStop);
}
inline Real Universe::getLeft() const{
  return myLeft;
}
inline Real Universe::getRight() const{
  return myRight;
}
inline Real Universe::getLx() const{
  return myLx;
}
inline Real Universe::getBottom() const{
  return myBottom;
}
inline Real Universe::getTop() const{
  return myTop;
}
inline Real Universe::getLy() const{
  return myLy;
}
inline long Universe::getCount() const{
  return myCount;
}
inline long Universe::getNStop() const{
  return myNStop;
}
inline Real Universe::getTStop() const{
  return myTStop;
}
inline Real Universe::getTime() const{
  return myTime;
}
inline Real Universe::getDt() const{
  return myDt;
}
inline long Universe::getNumberTotal() const{
  return myNumberTotal;
}
inline void Universe::setL0(Real l0) {
  myL0 = l0;
  myL1 = l0;
}

#endif // _UNIVERSE_H
