/*
 * springs.C
 *
 * Copyright (C) 1999-2000, Thierry Matthey <matthey@ii.uib.no>
 *
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

#ifdef HAVE_MPI
#include <mpi.h>
#endif
// Never ever include X11 before MPI !!!
#include <X11/Xlib.h>
#include <X11/keysym.h>

#include "mathutilities.h"
#include "global.h"
#include "display.h"
#include "Universe.h"
#include "StartWorld.h"
#include "Force.h"
#include "CellIndex.h"
#include "List.h"
#include "FastList.h"

//_________________________________________________________________ springs


void usage(char* name);
void parseCmdLine(int argc, char* argv[], char *in, char *out, int& savepic,
		  bool& flagTrace, int& width, int& height, bool& flagQuit, 
		  bool& flagNoGfx, bool& flagDump, bool& periodicX, bool& periodicY,
		  Universe::CommunicationType& com, Universe::PartitionType& part);

int main(int argc, char* argv[])
{
  char in[buffer_len];
  char out[buffer_len];

  XEvent report;
  KeySym key;
  Atom wm_quit;
  bool flagWait = false;
  bool flagTrace = false;
  bool flagQuit = false;
  bool flagNoGfx = false;
  bool quit = false;
  bool flagDump = true;
  int savepic;
  Universe::CommunicationType communication = Universe::CommunicationDistributed;
  Universe::PartitionType     partition     = Universe::PartitionStripe;
  periodic_boundary_x = true;
  periodic_boundary_y = true;
  periodic_boundary = periodic_boundary_x && periodic_boundary_y;


  int width = DEFAULT_WIDTH;
  int height = DEFAULT_HEIGHT;
  OutputWindow *window = NULL;

  //
  // Init MPI
  //
#ifdef HAVE_MPI
  MPI_Init(&argc,&argv);
  MPI_Comm_size(MPI_COMM_WORLD,&numProc);
  MPI_Comm_rank(MPI_COMM_WORLD,&myId);
  CellIndex::build_derived_type(&TYPE_CELLINDEX);
  Particle::build_derived_type(&TYPE_PARTICLE);
#endif

  //
  // Print I
  //
  if(myId == master){
    cerr << "Size of:"
	 << "Cell="<<sizeof(Cell)<<", "
	 << "Particle="<<sizeof(Particle)<<", "
	 << "CellIndex="<<sizeof(CellIndex)<<", "
	 << "bool="<<sizeof(bool)<<", "
	 << "char="<<sizeof(char)<<", "
	 << "short="<<sizeof(short)<<", "
	 << "int="<<sizeof(int)<<", "
	 << "long="<<sizeof(long)<<", "
	 << "long long="<<sizeof(long long)<<", "
	 << "float="<<sizeof(float)<<", "
	 << "double="<<sizeof(double)<<", "
	 << "long double="<<sizeof(long double)<<", "
	 << "Real="<<sizeof(Real)<<", "
	 << "List<Particle>="<<sizeof(List<Particle>)<<", "
	 << "FastList<Particle>="<<sizeof(FastList<Particle>)<<"\n\n";
  }

  //
  // Parse command line
  //
  parseCmdLine(argc, argv, in, out, savepic, flagTrace, width, height, flagQuit, flagNoGfx, flagDump, periodic_boundary_x, periodic_boundary_y, communication, partition);
  periodic_boundary = periodic_boundary_x && periodic_boundary_y;

  // Our debug file
  char filename[buffer_len];
  if(numProc > 1)
    sprintf(filename,"debug.%03d",myId);
  else
    sprintf(filename,"debug");

  debug.open(filename, ios::out);
  debug << "####### "<<myId<<","<<numProc<<"\n";
  debug.flush();
  debug.precision(16);
  debug.setf(ios::floatfield,ios::scientific);

  //
  // Print II
  //
  if(myId == master){
#ifdef HAVE_MPI
    cerr <<"Compiled with MPI, "<<numProc<<" process.\n";
#endif
#ifdef HAVE_ONE
    cerr <<"Compiled with MPI-2 one-sided communication.\n";
#endif
#ifdef HAVE_PULL
    cerr <<"Compiled with pull particles.\n";
#endif
#ifdef HAVE_EPOT
    cerr <<"Computing Epot.\n";
#endif
#ifdef HAVE_METIS
    cerr <<"Compiled with Metis.\n";
#endif
#ifdef HAVE_FORCE126
    cerr <<"Using hardcoded force computation, LJ 12-6.\n";
#endif
    cerr <<"Communication: "<<communication<<"\n"
	 <<"Partitioning : "<<partition<<"\n";
    if(periodic_boundary)
      cerr << "Periodic boundaries.\n";
    else if(periodic_boundary_x)
      cerr << "Periodic X-boundary.\n";
    else if(periodic_boundary_x)
      cerr << "Periodic Y-boundary.\n";
    cerr << "Mass     : "<<FROM_ATOMIC_UNIT_TO_KG<<" [Kg]\n"
	 << "Energy   : "<<FROM_EV_TO_J<<" [J]\n"
	 << "Length   : "<<FROM_ANGSTROM_TO_METER<<" [m]\n"
	 << "Time     : "<<FROM_ANGSTROM_TO_METER*sqrt(FROM_ATOMIC_UNIT_TO_KG/FROM_EV_TO_J)<<" [s], "<<FROM_ANGSTROM_TO_METER*sqrt(FROM_ATOMIC_UNIT_TO_KG/FROM_EV_TO_J)*1.0e9<<"[ns], "<<FROM_ANGSTROM_TO_METER*sqrt(FROM_ATOMIC_UNIT_TO_KG/FROM_EV_TO_J)*1.0e15<<"[fs]\n"
	 << "Velocity : "<<FROM_ANGSTROM_TO_METER/(FROM_ANGSTROM_TO_METER*sqrt(FROM_ATOMIC_UNIT_TO_KG/FROM_EV_TO_J))<<" [m/s]\n"
	 << "Force    : "<<FROM_EV_TO_J/FROM_ANGSTROM_TO_METER<<" [N]\n"
	 << "Pressure : "<<FROM_EV_TO_J/(FROM_ANGSTROM_TO_METER*FROM_ANGSTROM_TO_METER*FROM_ANGSTROM_TO_METER)<<" [Pa]\n";
    
  }

  // Create the universe
  Universe *universe = new Universe(new StartWorld(in, new Force()), communication, partition, flagDump);
  if (universe == NULL || universe->getNStop()<1){
    if(myId == master)
      cerr << "Could not create the universe.\n";
    delete universe;
    goto the_end;
  }

  // Open a display if wanted and possible
  if(universe->getNumberTotal() <= max_number_display){
    if(!flagNoGfx && myId == master){
      window = openDisplay(universe,width,height,true,savepic,argc,argv,in, out);
      if(window == NULL){
	flagNoGfx = true;
	if(myId == master)
	  cerr << "Could not open the display.\n";
      }
    }
  }
  else {
    if(!flagNoGfx && myId == master)
      cerr << "To many particles to open the display.\n";
    flagNoGfx = true;
  }

#ifdef HAVE_MPI
  MPI_Bcast(&flagNoGfx,1, MPI_BYTE, master,MPI_COMM_WORLD);
#endif


  if(flagNoGfx)
    flagQuit = true;

  // Exit & Close menuitem of the WM
  if(!flagNoGfx && myId == master){
    wm_quit = XInternAtom(window->display, "WM_DELETE_WINDOW", false);
    XSetWMProtocols(window->display, window->window, &wm_quit, 1);
  }

  // Display step 0
  if(myId == master){
    if(!flagNoGfx){
      writeDisplay(window,universe);
    }
    cerr << "Step: " << universe->getCount()<<" ("<<universe->getTime()<<")\n";
  }
  if(!flagNoGfx)
    universe->serveDisplay();


  // The main event loop
  while (!quit){
    
    if(!flagNoGfx && myId == master){

      if (flagTrace)
	flagWait = true;
      
      while ((XEventsQueued(window->display,QueuedAfterReading) > 0)||(universe->isDone()&&!quit)||flagWait){
	report.type = 0;
	XNextEvent(window->display, &report);
	
	switch (report.type){
	case ClientMessage:
	  if((unsigned)(report.xclient.data.l[0]) == wm_quit){
	    quit = true;
	    flagWait = false;
 	  }
	  break;

        case Expose: 
          while (XCheckTypedEvent(window->display, Expose, &report));
          break;
       
	case ConfigureNotify: // Window has been resized
	  if(resizeDisplay(window,report.xconfigure.width,report.xconfigure.height))
	    writeDisplay(window,universe);
          break;
	  
        case VisibilityNotify:
	  refreshDisplay(window);
	  break;
	  
        case ButtonPress:
          break;
	  
        case KeyPress:  // Check the keys
          key = XKeycodeToKeysym(window->display, report.xkey.keycode,0);
          switch (key){
          case XK_space:
            flagWait = (flagWait) ? false : true; 
            break;
          case XK_t:
            if (!flagTrace)
	      flagTrace = true;
	    else {
	      flagTrace = false;
	      flagWait = false;
	    }
            break;
          case XK_w:
            window->savepic |= SAVE_ONCE;
	    writeDisplay(window,universe);
            break;
          case XK_r:
            refreshDisplay(window);
	    break;
          case XK_h:
	    usage(argv[0]);
	    break;
          case XK_q: 
	    quit = true;
	    flagWait = false;
          default: // Throw away all other keys
            break;
          }
          break;
	  
        default:  // Throw away all other events
          break;
	}
      }
    }

    if(!flagNoGfx)
      universe->serveDisplay();


#ifdef HAVE_MPI
    MPI_Bcast(&quit,1, MPI_BYTE, master,MPI_COMM_WORLD);
    MPI_Bcast(&flagWait,1, MPI_BYTE, master,MPI_COMM_WORLD);
#endif

    if (!universe->isDone() && !flagWait && !quit){

      universe->next();
      if(master == myId)
	cerr << "Step: " << universe->getCount()<<" ("<<universe->getTime()<<")\n";

      if (universe->isDone()){
	if(master == myId)
	  cerr << "Done.\n";
	if (flagQuit)
          quit = true;
      }

      if(!flagNoGfx && myId == master)
	writeDisplay(window,universe);
    }


    if(!flagNoGfx)
      universe->serveDisplay();

  }
  
  /* exit gracefully */
  delete universe;
  closeDisplay(window);      

 the_end:
  debug.close();

#ifdef HAVE_MPI
  MPI_Type_free(&TYPE_CELLINDEX);
  MPI_Type_free(&TYPE_PARTICLE);
  MPI_Finalize();
#endif
  return(0);
}

void parseCmdLine(int argc, char* argv[], char* in, char *out, int& savepic,
		  bool& flagTrace, int& width, int& height, bool& flagQuit, 
		  bool& flagNoGfx, bool & flagDump, bool& periodicX, bool& periodicY,
		  Universe::CommunicationType& com, Universe::PartitionType& part){
  int cur = 1;

  savepic = 0;
  flagTrace = false;

  if ((argc >= 2 && (!strcmp(argv[1], "-h") || !strcmp(argv[1], "-help"))) || argc < 2) {
    if (myId == master)
      usage(argv[0]);

    debug.close();
#ifdef HAVE_MPI
    MPI_Type_free(&TYPE_CELLINDEX);
    MPI_Type_free(&TYPE_PARTICLE);
    MPI_Finalize();
#endif
    exit(0);
  }
  
  
  while (cur<argc && argv[cur][0]=='-') {

    if (!strcmp(argv[cur],"-width")) {
      width = atoi(argv[++cur]);
      cur++;
      continue;
    }

    if (!strcmp(argv[cur],"-x")){
      periodicX = true;
      cur++;
      continue;
    }
    if (!strcmp(argv[cur],"-nx")){
      periodicX = false;
      cur++;
      continue;
    }
    if (!strcmp(argv[cur],"-y")){
      periodicY = true;
      cur++;
      continue;
    }
    if (!strcmp(argv[cur],"-ny")){
      periodicY = false;
      cur++;
      continue;
    }

    if (!strcmp(argv[cur],"-com")) {
      long n = atoi(argv[++cur]);
      if(n<0 || n>4)
	n = 0;
      com = static_cast<Universe::CommunicationType>(n);
      cur++;
      continue;
    }

    if (!strcmp(argv[cur],"-par")) {
      long n = atoi(argv[++cur]);
      if(n<0 || n>3)
	n = 0;
      part = static_cast<Universe::PartitionType>(n);
      cur++;
      continue;
    }

    if (!strcmp(argv[cur],"-ps")){
      savepic |= SAVE_AS_PS;
      cur++;
      continue;
    }

    if (!strcmp(argv[cur],"-trace")){
      flagTrace = true;
      cur++;
      continue;
    }

    if (!strcmp(argv[cur],"-nogfx")){
      flagNoGfx = true;
      cur++;
      continue;
    }

    if (!strcmp(argv[cur],"-nodump")){
      flagDump = false;
      cur++;
      continue;
    }


    if (!strcmp(argv[cur],"-quit")){
      flagQuit = true;
      cur++;
      continue;
    }

    if (!strcmp(argv[cur],"-height")) {
      height = atoi(argv[++cur]);
      cur++;
      continue;
    }
    
    if (!strcmp(argv[cur],"-pbm")){
      savepic |= SAVE_AS_PBM;
      cur++;
      continue;
    }

    if (!strcmp(argv[cur],"-ppm")){
      savepic |= SAVE_AS_PPM;
      cur++;
      continue;
    }

    if (!strcmp(argv[cur],"-gif")){
      savepic |= SAVE_AS_GIF;
      cur++;
      continue;
    }

    if (!strcmp(argv[cur],"-all")){
      savepic |= SAVE_ALLWAYS;
      cur++;
      continue;
    }

    break;

  }

  if ((savepic == SAVE_ALLWAYS)|| (savepic == 0L))
    savepic |= SAVE_AS_PPM;
  
  in[0]  = 0;
  out[0] = 0;

  if (cur < argc) 
    strcpy(in,argv[cur++]);
  if (cur < argc)
    strcpy(out,argv[cur]);

  if((com == Universe::CommunicationSimple || com == Universe::CommunicationSequential) && part != Universe::PartitionStripe){
    part = Universe::PartitionStripe;
    if(myId == master)
      cerr << "Partitioning is not handled for the given communication.\n";
  }
#ifndef HAVE_METIS
  if(part == Universe::PartitionMetis){
    part = Universe::PartitionStripe;
    if(myId == master)
      cerr << "Have not METIS.\n";
  }
#endif
#if !defined(HAVE_ONE) || defined(_ABIN32)
  if(com == Universe::CommunicationOneSided || com == Universe::CommunicationOneSidedFast){
    com = Universe::CommunicationDistributed;
    if(myId == master)
      cerr << "Have not one-sided communications.\n";
  }
#endif

}

/*_________________________________________________________*/

void usage(char* name)
{
  cerr << " Usage: "<<name<<" [-width n][-height n][-nogfx][-trace][-pbm][-gif][-ppm][-ps][-all][-quit][-com n][-par n][in-file][out-file]\n"
       << "  where:\n"
       << "   -width  <n> : (optional) width of the window \n"
       << "   -height <n> : (optional) height of the window\n"
       << "   -pbm        : (optional) saving the pictures as PBM\n"
       << "   -gif        : (optional) saving the pictures as GIF\n"
       << "   -ppm        : (optional, default) saving the pictures as PPM\n"
       << "   -ps         : (optional) saves the pictures as PS\n"
       << "   -all        : (optional) saves every pictures\n"
       << "   -trace      : (optional) stops after each step\n"
       << "   -quit       : (optional) quit when finished\n"
       << "   -nogfx      : (optional) no display\n"
       << "   -nodump     : (optional) no dumps\n"
       << "   -x          : (optional) periodic X (default)\n"
       << "   -nx         : (optional) non-periodic X\n"
       << "   -y          : (optional) periodic Y (default)\n"
       << "   -ny         : (optional) non-periodic Y\n"
       << "   -com <n>    : (optional) communication ("
       << Universe::CommunicationDistributed<<"=distributed (default),\n"
       << "                                           "
       << Universe::CommunicationOneSided<<"=onesided,\n"
       << "                                           "
       << Universe::CommunicationSimple<<"=simple,\n"
       << "                                           "
       << Universe::CommunicationSequential<<"=sequential,\n"
       << "                                           "
       << Universe::CommunicationOneSidedFast<<"=onesidedfast)\n"
       << "   -par <n>    : (optional) partitioning ("
       << Universe::PartitionStripe<<"=stripe (default),\n"
       << "                                          "
       << Universe::PartitionMetis<<"=metis,\n"
       << "                                          "
       << Universe::PartitionChecker<<"=checker,\n"
       << "                                          "
       << Universe::PartitionRectangle<<"=rectangle)\n"
       << "   -h or -help : help\n"
       << "  keys:\n"
       << "   SPACE : stop/start\n"
       << "   h     : help\n"
       << "   q     : quit "<<name<<"\n"
       << "   r     : refresh display\n"
       << "   t     : trace on/off\n"
       << "   w     : screendump of the actuall screen\n";
}
